SystemOrganization addCategory: #'VMConstruction-Plugins-OSProcessPlugin'!

InterpreterPlugin subclass: #OSProcessPlugin
	instanceVariableNames: 'osprocessSandboxSecurity'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

!OSProcessPlugin commentStamp: '<historical>' prior: 0!
I provide pluggable primitive access to the operating system process in which the
Squeak VM is currently running. One or more of my subclasses may be translated
to produce platform specific plugins.!

!OSProcessPlugin class methodsFor: 'documentation' stamp: 'dtl 3/3/2002 13:14'!
about

	"OSProcessPlugin about"

	^ StringHolder new contents: self howToInstall; openLabel: 'about ', self asString! !

!OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 8/13/2003 21:58'!
canAccessSystem
	"True if an installed OSPP is available and can access the system"

	"OSProcessPlugin concreteClass canAccessSystem"

	^ self concreteClass primGetPid notNil
! !

!OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 8/31/2008 13:41'!
concreteClass

	"OSProcessPlugin concreteClass"

	| numericOsVersion |
	(self platformName = 'Win32') ifTrue: [^ Win32OSProcessPlugin].
	(self platformName = 'unix') ifTrue: [^ UnixOSProcessPlugin].
	numericOsVersion := self osVersion asInteger ifNil: [0].
	((self platformName = 'Mac OS') and: [numericOsVersion >= 1000])
		ifTrue: [^ UnixOSProcessPlugin concreteClass].
	self inform: 'no concrete class available for ', self platformName.
	^ nil
! !

!OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 9/15/2005 06:28'!
declareCVarsIn: cg

	"Additional declarations are done in my subclasses. This permits a meaningful
	module identification string to be written with the name of the actual subclass
	being generated."

	cg addHeaderFile: '"FilePlugin.h"'.
	cg addHeaderFile: '"SocketPlugin.h"
#define SESSIONIDENTIFIERTYPE int'.	"Data type for sessionID, should generate compiler warnings
									if someone changes the data type elsewhere in the VM."
	cg var: 'osprocessSandboxSecurity' declareC: 'int osprocessSandboxSecurity'.
! !

!OSProcessPlugin class methodsFor: 'version dependent' stamp: 'dtl 3/5/2005 11:51'!
getSystemAttribute: attributeID 
	"After Squeak version 3.6, #getSystemAttribute was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk getSystemAttribute: attributeID]) current getSystemAttribute: attributeID! !

!OSProcessPlugin class methodsFor: 'documentation' stamp: 'dtl 3/5/2005 12:06'!
howToInstall

	"OSProcessPlugin about"

	| installedPluginString |
	installedPluginString := self concreteClass installedModuleVersionString.
	^ (installedPluginString isNil
		ifTrue:
			['Your system does not have an installed plugin module. Your current configuration is ']
		ifFalse:
			[(self versionString = installedPluginString)
		ifTrue: ['Your system currently has an installed, up to date plugin module. Your current configuration is ']
		ifFalse: ['Your system currently has an installed plugin module, but its version does not match the OSProcessPlugin change set in your image, so you may wish to rebuild it. Your current configuration is ']]), self installedModule, '.

', self name, ' is a plugin written in Smalltalk. The complete source code for ', self name, ' for your platform is contained in class OSProcessPlugin and its subclasses. No external support code is required. It is intended to be translated to C and compiled as a shared object module (or linked directly with the VM if built as an internal plugin). ', self name, ' has a number of subclasses which allow platform specific methods to be generated for Unix, Linux and Windows platforms. The concrete subclass for your ', self platformName, ' Squeak is ', self concreteClass printString, '.

If OSProcess cannot connect to the operating system ("OSProcess accessor canAccessSystem" evaluates to false), then you need to build and install the ', self name,' plugin.

Use VMMaker to generate the C source code for ', self name, '. The correct ', self name, ' subclass will automatically appear in the list of available modules. VMMaker will create a C source code file called ', self concreteClass moduleName, '.c in the generated source code tree. Compile the plugin according to the instructions provided with your platform source distribution.

If ', self name, ' is built as an external plugin, the C source file will be compiled and linked into a shared module called UnixOSProcessPlugin.so (for Unix or Linux) or Win32OSProcessPlugin.dll (for Windows). If the plugin is built internally, it will be linked directly into the Squeak VM executable. The shared module file should be moved to a place where Squeak can find it (your Squeak home directory is fine). After restarting Squeak, OSProcess will begin functioning with access to the operating system.'! !

!OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 8/31/2008 13:25'!
idString

	^ 'D T Lewis'! !

!OSProcessPlugin class methodsFor: 'deprecated' stamp: 'dtl 8/8/2003 23:03'!
installedModule
	"Deprecated - this method name is misleading. Replaced as of OSP 3.1"

	^ self versionInformation! !

!OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:26'!
installedModuleName
	"Answer the module name string for the plugin which is actually running in the VM"

	"OSProcessPlugin installedModuleName"

	^ self concreteClass installedModuleName! !

!OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:25'!
installedModuleVersionString
	"Answer the version string for the plugin which is actually running in the VM"

	"OSProcessPlugin installedModuleVersionString"

	^ self concreteClass installedModuleVersionString! !

!OSProcessPlugin class methodsFor: 'version dependent' stamp: 'dtl 11/28/2010 10:26'!
osVersion
	"After Squeak version 3.6, #osVersion was moved to SmalltalkImage. Some
	versions of Pharo move this to OSPlatform and issue deprecation warnings
	about the other usages."

	"self osVersion"

	^ (((Smalltalk hasClassNamed: #OSPlatform)
			and: [(Smalltalk at: #OSPlatform)
					respondsTo: #osVersion])
		ifTrue: [Smalltalk at: #OSPlatform]
		ifFalse: [((Smalltalk classNamed: 'SmalltalkImage')
				ifNil: [^ Smalltalk osVersion]) current]) osVersion! !

!OSProcessPlugin class methodsFor: 'version dependent' stamp: 'dtl 11/28/2010 10:25'!
platformName 
	"After Squeak version 3.6, #platformName was moved to SmalltalkImage Some
	versions of Pharo move this to OSPlatform and issue deprecation warnings
	about the other usages."

	"self platformName"

	^ (((Smalltalk hasClassNamed: #OSPlatform)
			and: [(Smalltalk at: #OSPlatform)
					respondsTo: #platformName])
		ifTrue: [Smalltalk at: #OSPlatform]
		ifFalse: [((Smalltalk classNamed: 'SmalltalkImage')
				ifNil: [^ Smalltalk osVersion]) current]) platformName! !

!OSProcessPlugin class methodsFor: 'private - testing' stamp: 'dtl 6/12/2005 23:40'!
primFixPointersInArrayOfStrings: cStringArray withOffsets: offsetArray count: count
	"This primitive call exists only for purposes of testing the
	fixPointersInArrayOfStrings:withOffsets:count: method. I believe it to be
	reasonably machine and compiler independent, but have no way of verifying
	this on a variety of machines, so I'll leave this test method here in case
	someone runs into problems on other hardware or compilers. -dtl"

	"| a |
	a := OSProcess thisOSProcess envAsFlatArrayAndOffsets: UnixProcess env.
	OSProcessPlugin primFixPointersInArrayOfStrings: (a at: 1)
		withOffsets: (a at: 2)
		count: (a at: 2) size"

	<primitive: 'primitiveFixPointersInArrayOfStrings' module: 'UnixOSProcessPlugin'>
	^ nil! !

!OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 8/13/2003 21:56'!
primGetPid
	"Answer the OS process ID for the OS process in which I am currently executing."

	^ self subclassResponsibility! !

!OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 2/21/2002 15:57'!
shouldBeTranslated
	"Translate one of my subclasses instead."

	^ false! !

!OSProcessPlugin class methodsFor: 'version identification' stamp: 'dtl 3/5/2005 11:54'!
versionInformation

	"OSProcessPlugin versionInformation"

	| osp ospString |
	osp := Smalltalk at: #OSProcess ifAbsent: [].
	ospString := osp isNil
		ifTrue:
			[' (OSProcess not installed in this image)']
		ifFalse:
			[' installed with OSProcess version ', osp versionString].
	^ self vmVersion, ' with ', self concreteClass moduleName, ' version ',
		self versionString, ' using concrete class ', self concreteClass name,
		ospString, ', running on ', ((self getSystemAttribute: 1002) asString),
		' ', self platformName, ' VM with installed plugin ',
		(self canAccessSystem
			ifTrue:
				[self concreteClass installedModuleName isNil
					ifTrue:
						['(old version of plugin cannot identify itself)']
					ifFalse:
						[self concreteClass installedModuleName, ' version ',
						self installedModuleVersionString,
						((self versionString = self concreteClass installedModuleVersionString)
							ifTrue: [' (plugin matches class in image)']
							ifFalse: [' (plugin does not match class in image)'])]]
			ifFalse:
				['(OSProcessPlugin installed module not accessible)'])

! !

!OSProcessPlugin class methodsFor: 'version identification' stamp: 'dtl 1/20/2012 20:28'!
versionString
	"Version numbering is independent of OSProcess class>>versionString, and should be
	changed only when the plugin changes. Version numbers will begin arbitrarily
	with 2.1 as of OSProcess 2.7."

	"OSProcessPlugin versionString"

	^ self new versionString
! !

!OSProcessPlugin class methodsFor: 'version dependent' stamp: 'dtl 3/5/2005 11:52'!
vmVersion
	"After Squeak version 3.6, #vmVersion was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk vmVersion]) current vmVersion! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl (auto pragmas dtl 2010-11-28) 2/21/2002 16:30'!
cStringFromString: aString
	"Answer a new null-terminated C string copied from aString. The C string
	is allocated from the C runtime heap. See transientCStringFromString for
	a version which allocates from object memory.
	Caution: This may invoke the garbage collector."

	| len sPtr cString |
	<returnTypeC: 'char *'>
	<var: 'sPtr' declareC: 'char *sPtr'>
	<var: 'cString' declareC: 'char *cString'>
	sPtr := interpreterProxy arrayValueOf: aString.
	len := interpreterProxy sizeOfSTArrayFromCPrimitive: sPtr.
	cString := self callocWrapper: len + 1 size: 1.			"Space for a null terminated C string."
	self cCode: '(char *) strncpy (cString, sPtr, len)'.	"Copy the string."
	^ cString
! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl (auto pragmas dtl 2010-11-28) 2/21/2002 16:30'!
callocWrapper: count size: objectSize
	"Using malloc() and calloc() is something I would like to avoid, since it is
	likely to cause problems some time in the future if somebody redesigns
	object memory allocation. This wrapper just makes it easy to find senders
	of calloc() in my code. -dtl"

	<returnTypeC: 'void *'>
	^ self cCode: 'calloc(count, objectSize)'! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 5/8/2011 22:13'!
copyBytesFrom: charArray1 to: charArray2 length: len

	<var: 'charArray1' type: 'void *'>
	<var: 'charArray2' type: 'void *'>
	self cCode: 'memcpy(charArray2, charArray1, len)'
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 09:22'!
fileHandleFrom: sqFileStructByteArray
	"Answer a file handle from a SQFile structure. On most platforms, this
	will be a (FILE *). On Win32, it is a HANDLE."

	| sqFile |
	<returnTypeC: 'FILEHANDLETYPE'>
	<var: 'sqFile' declareC: 'SQFile *sqFile'>
	sqFile := interpreterProxy arrayValueOf: sqFileStructByteArray.
	^ self cCode: 'sqFile->file' inSmalltalk: [0]
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 3/24/2015 21:44'!
fileRecordSize
	"Answer the size of a SQFile data structure in bytes."

	^ self sizeof: #SQFile! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 11:05'!
fileValueOf: anSQFileRecord
	"Return a pointer to the first byte of of the SQFile data structure file record within
	anSQFileRecord, which is expected to be a ByteArray of size self>>fileRecordSize."

	<returnTypeC: 'SQFile *'>
	^ interpreterProxy arrayValueOf: anSQFileRecord
! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl (auto pragmas dtl 2010-11-28) 7/12/2005 07:05'!
fixPointersInArrayOfStrings: flattenedArrayOfStrings withOffsets: offsetArray count: count
	"Use the address offsets in offsetArray to fix up the pointers in cStringArray.
	The result is a C array of pointers to char, used for argv and env vectors."

	| ptr idx |
	<var: 'flattenedArrayOfStrings' declareC: 'char *flattenedArrayOfStrings'>
	<var: 'offsetArray' declareC: 'sqInt *offsetArray'>
	<var: 'ptr' declareC: 'char **ptr'>
	ptr := self cCoerce: flattenedArrayOfStrings to: 'char **'.
	idx := 0.
	[idx < count]
		whileTrue:
			[ptr
				at: idx
				put: (flattenedArrayOfStrings + (interpreterProxy integerValueOf: (offsetArray at: idx))).
			idx := idx + 1]! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl 9/1/2008 10:35'!
getThisSessionIdentifier

	^ interpreterProxy getThisSessionID
! !

!OSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/4/2006 19:43'!
initialiseModule

	<export: true>
	osprocessSandboxSecurity := -1.
	self initializeModuleForPlatform.
	^ true
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 12:45'!
isNonNullSQFile: objectPointer
	"Check for the common failure mode of a SQFile record with all zeros."

	| idx sqFileBytes  |
	<var: 'sqFileBytes' type: 'unsigned char *'>
	sqFileBytes := interpreterProxy arrayValueOf: objectPointer.
	idx := 0.
	[idx < self fileRecordSize]
		whileTrue:
			[((sqFileBytes at: idx) ~= 0) ifTrue: [^ true].
			idx := idx + 1].
	^ false
! !

!OSProcessPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 12:04'!
isNullSQSocket: objectPointer
	"Check for the common failure mode of a SQSocket record with all zeros."

	| idx sqSocketBytes |
	<var: 'sqSocketBytes' type: 'unsigned char *'>
	sqSocketBytes := interpreterProxy arrayValueOf: objectPointer.
	idx := 0.
	[idx < self socketRecordSize]
		whileTrue:
			[((sqSocketBytes at: idx) ~= 0) ifTrue: [^ false].
			idx := idx + 1].
	^ true
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 4/29/2006 12:45'!
isSQFileObject: objectPointer 
	"Answer true if objectPointer appears to be a valid SQFile ByteArray.
	This check is appropriate if objectPointer has been passed as a parameter
	to a primitive, and is expected to represent a valid file reference."

	^ (((interpreterProxy isBytes: objectPointer)
		and: [(interpreterProxy byteSizeOf: objectPointer) = self fileRecordSize])
		and: [self isValidFileSession: objectPointer])
		and: [self isNonNullSQFile: objectPointer]! !

!OSProcessPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 4/29/2006 11:55'!
isSQSocketObject: objectPointer
	"Answer true if objectPointer appears to be a valid SQSocket ByteArray. This check
	is appropriate if objectPointer has been passed as a parameter to a primitive, and
	is expected to represent a valid socket reference."

	^ (((interpreterProxy isBytes: objectPointer)
		and: [(interpreterProxy byteSizeOf: objectPointer) = self socketRecordSize]))
		and: [(self isNullSQSocket: objectPointer) not]
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 4/29/2006 11:25'!
isValidFileSession: objectPointer
	"Answer true if the file session matches the current interpreter session identifier."

	 ^ self getThisSessionIdentifier =
		(self sessionIdentifierFromSqFile: (interpreterProxy arrayValueOf: objectPointer))
! !

!OSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl (auto pragmas dtl 2010-11-28) 6/7/2008 14:53'!
moduleUnloaded: aModuleName
	"The module with the given name was just unloaded.
	Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 2/21/2002 16:35'!
newSQFileByteArray
	"Answer a new ByteArray sized to contain a SQFile data structure."

	^ interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: self fileRecordSize
! !

!OSProcessPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 2/21/2002 16:36'!
newSQSocketByteArray
	"Answer a new ByteArray sized to contain a SQSocket data structure."

	^ interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: self socketRecordSize
! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl 6/3/2013 18:32'!
nullTerminatedStringFromString: aString
	"Answer a new null-terminated string copied from aString. The result is
	a string of size one greated than aString, with its last character set to zero.
	May cause garbage collection."

	| len sourceString stringPtr newString cString |
	<var: 'stringPtr' declareC: 'char *stringPtr'>
	<var: 'cString' declareC: 'char *cString'>
	len := interpreterProxy sizeOfSTArrayFromCPrimitive: (interpreterProxy arrayValueOf: aString).
	interpreterProxy pushRemappableOop: aString.
	newString := interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len + 1.
	sourceString := interpreterProxy popRemappableOop.
	stringPtr := interpreterProxy arrayValueOf: sourceString.
	cString := interpreterProxy arrayValueOf: newString.	"Point to the actual C string."
	self cCode: '(char *)strncpy(cString, stringPtr, len)'.		"Make a copy of the string."
	cString at: (len) put: 0.									"Null terminate the C string."
	^ sourceString
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/2/2002 18:44'!
pointerFrom: aByteArray
	"Answer the pointer represented by aByteArray."

	| ptr pointerUnion idx |
	<returnTypeC: 'void *'>
	<var: 'ptr' declareC: 'unsigned char *ptr'>
	<var: 'pointerUnion' declareC: 'union {void *address; unsigned char bytes[sizeof(void *)];} pointerUnion'>
	((interpreterProxy isBytes: aByteArray) and:
		[(interpreterProxy stSizeOf: aByteArray) = self sizeOfPointer])
		ifFalse: [^ nil].
	ptr := interpreterProxy arrayValueOf: aByteArray.
	idx := 0.
	[idx < self sizeOfPointer]	whileTrue:
		[self cCode: 'pointerUnion.bytes[idx] = ptr[idx]'.
		idx := idx + 1].
	^ self cCode: 'pointerUnion.address' inSmalltalk: [pointerUnion]
! !

!OSProcessPlugin methodsFor: 'primitives - pipes' stamp: 'dtl (auto pragmas dtl 2010-11-28) 6/13/2005 00:21'!
primitiveCreatePipe
	"Create a pipe, and answer an array of two file handles for the pipe writer and reader.
	The readerIOStream and writerIOStream variables represent the low level pipe streams,
	which will be of type (FILE *) or HANDLE, depending on what the FilePlugin support
	code is using to represent file streams. FILEHANDLETYPE is defined in my subclasses
	in the #declareCVarsIn: class method."

	| reader writer readerPtr writerPtr readerIOStream writerIOStream
	readerIOStreamPtr writerIOStreamPtr thisSession arrayResult |
	<export: true>
	<var: 'readerIOStream' type: 'FILEHANDLETYPE'>
	<var: 'writerIOStream' type: 'FILEHANDLETYPE'>
	<var: 'readerIOStreamPtr' declareC: 'FILEHANDLETYPE *readerIOStreamPtr'>
	<var: 'writerIOStreamPtr' declareC: 'FILEHANDLETYPE *writerIOStreamPtr'>
	<var: 'readerPtr' declareC: 'SQFile *readerPtr'>
	<var: 'writerPtr' declareC: 'SQFile *writerPtr'>
	<var: 'arrayPtr' declareC: 'int *arrayPtr'>
	<var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'>
	thisSession := self getThisSessionIdentifier.

	"Create the anonymous OS pipe"
	self cCode: 'readerIOStreamPtr = &readerIOStream'.
	self cCode: 'writerIOStreamPtr = &writerIOStream'.
	(self createPipeForReader: readerIOStreamPtr writer: writerIOStreamPtr)
		ifFalse: [^ interpreterProxy primitiveFail].

	"Map the writer end to a SQFile data structure"
	writer := self newSQFileByteArray.
	writerPtr := self fileValueOf: writer.
	self cCode: 'writerPtr->file = writerIOStream'.
	self cCode: 'writerPtr->sessionID = thisSession'.
	self cCode: 'writerPtr->writable = 1'.
	self cCode: 'writerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: writer.

	"Map the reader end to a SQFile data structure"
	reader := self newSQFileByteArray.
	readerPtr := self fileValueOf: reader.
	self cCode: 'readerPtr->file = readerIOStream'.	
	self cCode: 'readerPtr->sessionID = thisSession'.
	self cCode: 'readerPtr->writable = 0'.
	self cCode: 'readerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: reader.

	"Answer an array of two SQFile pipe ends"
	arrayResult := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2.
	interpreterProxy stObject: arrayResult at: 1 put: interpreterProxy popRemappableOop.	"reader"
	interpreterProxy stObject: arrayResult at: 2 put: interpreterProxy popRemappableOop. "writer"
	interpreterProxy pop: 1; push: arrayResult
! !

!OSProcessPlugin methodsFor: 'primitives - pipes' stamp: 'dtl (auto pragmas dtl 2010-11-28) 6/13/2005 00:21'!
primitiveCreatePipeWithSessionIdentifier
	"Create a pipe, and answer an array of two file handles for the pipe writer and reader.
	The session identifier is passed as the parameter to this primitive. Use this variant
	if the session identifier is not available directly in the VM (as may be the case if
	it is not possible to link from this plugin to a variable elsewhere in the VM).
	The readerIOStream and writerIOStream variables represent the low level pipe streams,
	which will be of type (FILE *) or HANDLE, depending on what the FilePlugin support
	code is using to represent file streams. FILEHANDLETYPE is defined in my subclasses
	in the #declareCVarsIn: class method."

	| reader writer readerPtr writerPtr readerIOStream writerIOStream
	readerIOStreamPtr writerIOStreamPtr thisSession arrayResult |
	<export: true>
	<var: 'readerIOStream' type: 'FILEHANDLETYPE'>
	<var: 'writerIOStream' type: 'FILEHANDLETYPE'>
	<var: 'readerIOStreamPtr' declareC: 'FILEHANDLETYPE *readerIOStreamPtr'>
	<var: 'writerIOStreamPtr' declareC: 'FILEHANDLETYPE *writerIOStreamPtr'>
	<var: 'readerPtr' declareC: 'SQFile *readerPtr'>
	<var: 'writerPtr' declareC: 'SQFile *writerPtr'>
	<var: 'arrayPtr' declareC: 'int *arrayPtr'>
	<var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'>
	thisSession := self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).

	"Create the anonymous OS pipe"
	self cCode: 'readerIOStreamPtr = &readerIOStream'.
	self cCode: 'writerIOStreamPtr = &writerIOStream'.
	(self createPipeForReader: readerIOStreamPtr writer: writerIOStreamPtr)
		ifFalse: [^ interpreterProxy primitiveFail].

	"Map the writer end to a SQFile data structure"
	writer := self newSQFileByteArray.
	writerPtr := self fileValueOf: writer.
	self cCode: 'writerPtr->file = writerIOStream'.
	self cCode: 'writerPtr->sessionID = thisSession'.
	self cCode: 'writerPtr->writable = 1'.
	self cCode: 'writerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: writer.

	"Map the reader end to a SQFile data structure"
	reader := self newSQFileByteArray.
	readerPtr := self fileValueOf: reader.
	self cCode: 'readerPtr->file = readerIOStream'.	
	self cCode: 'readerPtr->sessionID = thisSession'.
	self cCode: 'readerPtr->writable = 0'.
	self cCode: 'readerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: reader.

	"Answer an array of two SQFile pipe ends"
	arrayResult := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2.
	interpreterProxy stObject: arrayResult at: 1 put: interpreterProxy popRemappableOop.	"reader"
	interpreterProxy stObject: arrayResult at: 2 put: interpreterProxy popRemappableOop. "writer"
	interpreterProxy pop: 2; push: arrayResult
! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl (auto pragmas dtl 2010-11-28) 7/12/2005 22:57'!
primitiveFixPointersInArrayOfStrings
	"This primitive exists only for purposes of testing the
	fixPointersInArrayOfStrings:withOffsets:count: method. I believe it to be
	reasonably machine and compiler independent, but have no way of verifying
	this on a variety of machines, so I'll leave this test method here in case
	someone runs into problems on other hardware or compilers. -dtl"

	| count offsetArray offsets cStringArray flattenedArrayOfStrings |
	<export: true>
	<var: 'offsets' declareC: 'sqInt *offsets'>
	<var: 'flattenedArrayOfStrings' declareC: 'char *flattenedArrayOfStrings'>
	count := interpreterProxy stackIntegerValue: 0.
	offsetArray := interpreterProxy stackObjectValue: 1.
	cStringArray := interpreterProxy stackObjectValue: 2.
	offsets := interpreterProxy firstIndexableField: offsetArray.
	flattenedArrayOfStrings := interpreterProxy arrayValueOf: cStringArray.
	self
		fixPointersInArrayOfStrings: flattenedArrayOfStrings
		withOffsets: offsets
		count: count.
	interpreterProxy pop: 4; push: cStringArray! !

!OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 8/21/2011 22:48'!
primitiveGetSession
	"Answer the unique session identifier for this Smalltalk instance running in this
	OS process. The C integer value is coerced into a Smalltalk ByteArray to preserve
	the full range of possible values."

	| sessionOop sessionByteArrayPointer thisSessionID sessionIDSize |
	<export: true>
	<var: 'sessionByteArrayPointer' declareC: 'unsigned char *sessionByteArrayPointer'>
	<var: 'copyByteArrayPointer' declareC: 'unsigned char *copyByteArrayPointer'>
	<var: 'thisSessionID' type: 'SESSIONIDENTIFIERTYPE'>
	thisSessionID := self getThisSessionIdentifier.
	sessionIDSize := self cCode: 'sizeof(thisSessionID)' inSmalltalk: [Smalltalk wordSize].
	sessionOop := interpreterProxy
		instantiateClass: interpreterProxy
		classByteArray
		indexableSize: sessionIDSize.
	sessionByteArrayPointer := interpreterProxy arrayValueOf: sessionOop.
	thisSessionID isNil ifTrue:
		[^ interpreterProxy primitiveFail].	"The session ID is not permitted to be zero"
	self copyBytesFrom: (self cCode: '(unsigned char *)&thisSessionID' inSmalltalk: [thisSessionID])
		to: sessionByteArrayPointer
		length: sessionIDSize.
	interpreterProxy pop: 1; push: sessionOop
! !

!OSProcessPlugin methodsFor: 'primitives - version testing' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 17:50'!
primitiveModuleName
	"Answer a string containing the module name string for this plugin."

	<export: true>
	interpreterProxy pop: 1 thenPush: (self stringFromCString: self getModuleName)
! !

!OSProcessPlugin methodsFor: 'primitives - machine specific' stamp: 'dtl (auto pragmas dtl 2010-11-28) 7/14/2003 06:32'!
primitiveSizeOfInt
	"Size in bytes of an integer, for this C compiler on this machine."

	<export: true>
	interpreterProxy pop: 1; pushInteger: self sizeOfInt! !

!OSProcessPlugin methodsFor: 'primitives - machine specific' stamp: 'dtl (auto pragmas dtl 2010-11-28) 7/14/2003 06:33'!
primitiveSizeOfPointer
	"Size in bytes of a void pointer, for this C compiler on this machine."

	<export: true>
	interpreterProxy pop: 1; pushInteger: self sizeOfPointer! !

!OSProcessPlugin methodsFor: 'primitives - version testing' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/1/2002 16:56'!
primitiveVersionString
	"Answer a string containing the version string for this plugin."

	<export: true>
	interpreterProxy pop: 1; push: (self stringFromCString: self versionString)
! !

!OSProcessPlugin methodsFor: 'private - security' stamp: 'dtl 3/3/2002 08:44'!
sandboxSecurity
	"Answer 1 if running in secure mode, else 0. The osprocessSandboxSecurity
	variable is initialized to -1. On the first call to this method, set its value to
	either 0 (user has full access to the plugin) or 1 (user is not permitted to do
	dangerous things)."

	(osprocessSandboxSecurity < 0) ifTrue: [osprocessSandboxSecurity := self securityHeurisitic].
	^ osprocessSandboxSecurity
! !

!OSProcessPlugin methodsFor: 'private - security' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/18/2005 12:37'!
securityHeurisitic
	"Answer 0 to permit full access to OSProcess functions, or 1 if access should be
	restricted for dangerous functions. The rules are:
		- If the security plugin is not present, grant full access
		- If the security plugin can be loaded, restrict access unless user has all
		  of secCanWriteImage, secHasFileAccess and secHasSocketAccess"

	"FIXME: This function has not been tested. -dtl"

	| sCWIfn sHFAfn sHSAfn canWriteImage hasFileAccess hasSocketAccess |
	"If the security plugin can be loaded, use it to check. If not, assume it's ok"

	<var: 'sCWIfn' declareC: 'void (*sCWIfn)(void)'>
	<var: 'sHFAfn' declareC: 'void (*sHFAfn)(void)'>
	<var: 'sHSAfn' declareC: 'void (*sHSAfn)(void)'>

	sCWIfn := interpreterProxy ioLoadFunction: 'secCanWriteImage' From: 'SecurityPlugin'.
	(sCWIfn = 0) ifTrue: [^ 0 "Cannot find plugin, so grant default full access"].
	canWriteImage := self cCode: ' ((int (*) (void)) sCWIfn)()'
				inSmalltalk: [1].
	sHFAfn := interpreterProxy ioLoadFunction: 'secHasFileAccess' From: 'SecurityPlugin'.
	(sHFAfn = 0) ifTrue: [^ 0 "Cannot find plugin, so grant default full access"].
	hasFileAccess := self cCode: ' ((int (*) (void)) sHFAfn)()'
				inSmalltalk: [1].
	sHSAfn := interpreterProxy ioLoadFunction: 'secHasSocketAccess' From: 'SecurityPlugin'.
	(sHSAfn = 0) ifTrue: [^ 0 "Cannot find plugin, so grant default full access"].
	hasSocketAccess := self cCode: ' ((int (*) (void)) sHSAfn)()'
				inSmalltalk: [1].
	((canWriteImage and: [hasFileAccess]) and: [hasSocketAccess])
		ifTrue: [^ 0 "Grant full access"]
		ifFalse: [^ 1 "Restricted access"]
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/3/2002 10:55'!
sessionIdentifierFrom: aByteArray
	"Answer a session ID represented by aByteArray. The session ID is used in
	the SQFile structure. If that data structure changes, we should see compiler
	warnings about type mismatch with SESSIONIDENTIFIERTYPE."

	| session sessionUnion idx |
	<returnTypeC: 'SESSIONIDENTIFIERTYPE'>
	<var: 'session' declareC: 'unsigned char *session'>
	<var: 'sessionUnion' declareC: 'union {SESSIONIDENTIFIERTYPE session; unsigned char bytes[sizeof(SESSIONIDENTIFIERTYPE)];} sessionUnion'>
	((interpreterProxy isBytes: aByteArray) and:
		[(interpreterProxy stSizeOf: aByteArray) = self sizeOfSession])
		ifFalse: [^ nil].
	session := interpreterProxy arrayValueOf: aByteArray.
	idx := 0.
	[idx < self sizeOfSession]	whileTrue:
		[self cCode: 'sessionUnion.bytes[idx] = session[idx]'.
		idx := idx + 1].
	^ self cCode: 'sessionUnion.session' inSmalltalk: [sessionUnion]
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/26/2006 20:57'!
sessionIdentifierFromSqFile: sqFile
	"Answer the session identifier from a SQFile structure. For a valid file
	reference, this identifier will match the session identifier supplied by
	the interpreter."

	<inline: true>
	<returnTypeC: 'SESSIONIDENTIFIERTYPE'>
	<var: 'sqFile' type: 'SQFile *'>
	^ self cCode: 'sqFile->sessionID' inSmalltalk: [-1]
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 7/14/2003 06:31'!
sizeOfInt
	"Size in bytes of an integer, for this C compiler on this machine."

	^ self cCode: 'sizeof(int)' inSmalltalk: [4]
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 7/14/2003 06:32'!
sizeOfPointer
	"Size in bytes of a void pointer, for this C compiler on this machine."

	^ self cCode: 'sizeof(void *)' inSmalltalk: [4].
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 3/3/2002 10:53'!
sizeOfSession
	"Size of a SESSIONIDENTIFIERTYPE. Should match usage in the SQFile data structure,
	otherwise we should get compiler warnings."

	^ self cCode: 'sizeof(SESSIONIDENTIFIERTYPE)' inSmalltalk: [4]
! !

!OSProcessPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 13:14'!
socketDescriptorFrom: sqSocketOop
	"Answer the OS file descriptor, an integer value, from a SQSocket data structure,
	or answer -1 if unable to obtain the file descriptor (probably due to receiving
	an incorrect type of object as aFileHandle).

	Warning: The first element of privateSocketStruct happens to be the Unix file
	number of the socket. See sqUnixSocket.c for the definition. This method takes
	advantage of this, and will break if anyone ever redefines the data structure."

	| sqSocket privateSocketStruct |
	<returnTypeC: 'int'>
	<var: 'sqSocket' declareC: 'SQSocket *sqSocket'>
	<var: 'privateSocketStruct' declareC: 'void *privateSocketStruct'>
	sqSocket := self socketValueOf: sqSocketOop.
	privateSocketStruct := self cCode: 'sqSocket->privateSocketPtr'.
	privateSocketStruct = 0
		ifTrue: [^ -1].	"Something is wrong, bail out before we core dump."
	^ self cCode: '* (int *) privateSocketStruct' inSmalltalk: [-1].
! !

!OSProcessPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 3/24/2015 21:44'!
socketRecordSize
	"Answer the size of a SQSocket data structure in bytes."

	^ self sizeof: #SQSocket! !

!OSProcessPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 11:48'!
socketValueOf: anSQSocketRecord
	"Return a pointer to the first byte of of the SQsocket data structure socket record within
	anSQSocketRecord, which is expected to be a ByteArray of size self>>socketRecordSize."

	<returnTypeC: 'SQSocket *'>
	^ interpreterProxy arrayValueOf: anSQSocketRecord
! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl 9/11/2011 08:16'!
stringFromCString: aCString
	"Answer a new String copied from a null-terminated C string.
	Caution: This may invoke the garbage collector."

	| len newString p |
	<var: 'aCString' type: 'const char *'>
	<var: 'p' type: 'void *'>
	len := self
		cCode: 'strlen(aCString)'
		inSmalltalk: [aCString size].
	newString := interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len.
	p := interpreterProxy arrayValueOf: newString.
	self cCode: 'strncpy(p, aCString, len)'
		inSmalltalk: [ | s |
			s := interpreterProxy firstIndexableField: newString.
			(1 to: len) do: [:i | s at: i - 1 put: (aCString at: i) asciiValue]].
	^ newString

! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/3/2003 14:55'!
transientCStringFromString: aString
	"Answer a new null-terminated C string copied from aString.
	The string is allocated in object memory, and will be moved
	without warning by the garbage collector. Any C pointer
	reference the the result is valid only until the garbage
	collector next runs. Therefore, this method should only be used
	within a single primitive in a section of code in which the
	garbage collector is guaranteed not to run. Note also that
	this method may itself invoke the garbage collector prior
	to allocating the new C string.

	Warning: The result of this method will be invalidated by the
	next garbage collection, including a GC triggered by creation
	of a new object within a primitive. Do not call this method
	twice to obtain two string pointers."

	| len stringPtr newString cString |
	<returnTypeC: 'char *'>
	<var: 'stringPtr' declareC: 'char *stringPtr'>
	<var: 'cString' declareC: 'char *cString'>
	len := interpreterProxy sizeOfSTArrayFromCPrimitive: (interpreterProxy arrayValueOf: aString).
	"Allocate space for a null terminated C string."
	interpreterProxy pushRemappableOop: aString.
	newString := interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len + 1.
	stringPtr := interpreterProxy arrayValueOf: interpreterProxy popRemappableOop.
	cString := interpreterProxy arrayValueOf: newString.		"Point to the actual C string."
	self cCode: '(char *)strncpy(cString, stringPtr, len)'.		"Make a copy of the string."
	cString at: (len) put: 0.									"Null terminate the C string."
	^ cString
! !

!OSProcessPlugin methodsFor: 'version identification' stamp: 'dtl 2/22/2017 20:11'!
versionString
	"Answer a string identifying the version level for this plugin."

	| version |
	<returnTypeC: 'char *'>
	<var: 'version' declareC: 'static char version[]= "4.6.2"'>
	^ self cCode: 'version' inSmalltalk: ['4.6.2']
! !

OSProcessPlugin subclass: #UnixOSProcessPlugin
	instanceVariableNames: 'sigChldSemaIndex originalSigHandlers semaIndices vmThread sigNumToSend pidArray pidCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

!UnixOSProcessPlugin commentStamp: '<historical>' prior: 0!
I provide pluggable primitive access to the Unix operating system process in which the
Squeak VM is currently running.!

!UnixOSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 2/22/2017 20:10'!
declareCVarsIn: cg

	"Additional declarations are done in my subclasses. This permits a meaningful
	module identification string to be written with the name of the actual subclass
	being generated."

	cg addHeaderFile: '<sys/types.h>
/* ', self idString, ' - ', self moduleName, '.c translated from class
   ', self name, ' of OSProcessPlugin version ', self versionString, ' */'.
	cg addHeaderFile: '<sys/time.h>'.
	cg addHeaderFile: '<sys/wait.h>'.
	cg addHeaderFile: '<signal.h>'.
	cg addHeaderFile: '<unistd.h>'.
	cg addHeaderFile: '<stdlib.h>'.
	cg addHeaderFile: '<fcntl.h>'.
	cg addHeaderFile: '<signal.h>'.
	cg addHeaderFile: '<pthread.h>'.
	cg addHeaderFile: '<sys/stat.h>'.	"For stat(2) call in primitiveFileProtectionMask"
	cg addHeaderFile: '<errno.h>'.
	cg addHeaderFile: '"config.h"
#define FILEHANDLETYPE FILE *  /* the type of low level stream to be used in a struct SQFile */'.
	cg addHeaderFile: '"sqaio.h"'.

	"signal handler arrays use one-based indexing with the zero position unused"
	cg var: 'originalSigHandlers' declareC: 'void *originalSigHandlers[NSIG + 1]'.
	cg var: 'semaIndices' declareC: 'unsigned char semaIndices[NSIG + 1]'.

	cg var: 'exitFunction' declareC: 'void (*exitFunction)(void) = NULL'.
	cg var: 'sigNumToSend' declareC: 'int sigNumToSend = SIGTERM'.
	cg var: 'pidArray' declareC: 'pid_t *pidArray = NULL'.
	cg var: 'vmThread' type: 'pthread_t'

! !

!UnixOSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:05'!
installedModuleName
	"Answer the module name string for the plugin which is actually running in the VM"

	"UnixOSProcessPlugin installedModuleName"

	<primitive: 'primitiveModuleName' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:24'!
installedModuleVersionString
	"Answer the version string for the plugin which is actually running in the VM"

	"UnixOSProcessPlugin installedModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 7/26/2010 22:39'!
isSuitablePluginForPlatform: platformName
	"Answer true is this is an instance of the class which is responsible for
	representing the named platform. Newer versions of VMMaker use this
	mechanism"

	^ platformName = 'unix'
		or: [platformName = 'Mac OS']! !

!UnixOSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 7/22/2000 17:15'!
moduleName

	^ 'UnixOSProcessPlugin'! !

!UnixOSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 8/13/2003 21:54'!
primGetPid
	"Answer the OS process ID for the OS process in which I am currently executing."

	<primitive: 'primitiveGetPid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 9/11/2011 09:36'!
requiredMethodNames
	"return the list of method names that should be retained for export or other support reasons"

	^{ #handleSignal: . #sendSignalToPids }! !

!UnixOSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 7/26/2010 22:29'!
shouldBeTranslated

	^ true
! !

!UnixOSProcessPlugin methodsFor: 'private - pipes' stamp: 'dtl (auto pragmas dtl 2010-11-28) 1/31/2004 18:47'!
createPipeForReader: readerIOStreamPtr writer: writerIOStreamPtr
	"Create a pipe and populate the readerIOStream and writerIOStream variables.
	The SIGPIPE handler must have been set before creating the pipe. Answer true for
	success, else false."

	| filedes |
	<var: 'readerIOStreamPtr' type: 'FILEHANDLETYPE *'>
	<var: 'writerIOStreamPtr' type: 'FILEHANDLETYPE *'>
	<var: 'filedes' declareC: 'int filedes[2]'>
	((self pipe: filedes) = -1) "Translates to a pipe() system call"
		ifTrue:
			[^ false]
		ifFalse:
			[self cCode: '*writerIOStreamPtr= (FILE *) fdopen (filedes[1], "a")'.
			self cCode: '*readerIOStreamPtr= (FILE *) fdopen (filedes[0], "r")'.
			^ true]
! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 2/24/2001 11:39'!
descriptorTableSize
	"Answer the size of the file descriptor table for a process. I am not sure of the most portable
	way to do this. If this implementation does not work on your Unix platform, try changing
	it to answer the value of FOPEN:=MAX, which will hopefully be defined in stdio.h. If
	all else fails, just hard code it to answer 20, which would be safe for any Unix."

	^ self cCode: 'getdtablesize()' inSmalltalk: [20]
! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl (auto pragmas dtl 2010-11-28) 1/30/2003 22:42'!
dupToStdErr: anSQFileDataStructure
	"Dup a file descriptor to allow it to be attached as the standard error when we
	exec() a new executable. This is Unix specific, in that it assumes that file descriptor
	0 is stdin, 1 is stdout, and 2 is stderr. The dup2() call is used to copy the open file
	descriptors into file descriptors 0, 1 and 2 so that the image which we execute will
	use them as stdin, stdout, and stderr."

	| filenoToDup |
	<returnTypeC: 'void'>
	filenoToDup := self fileDescriptorFrom: anSQFileDataStructure.
	(filenoToDup < 0) ifFalse: [
		(filenoToDup = 2) ifFalse: [
			self cCode: 'fflush(stderr)'.
			self cCode: 'dup2(filenoToDup, 2)' ]]! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl (auto pragmas dtl 2010-11-28) 1/30/2003 22:42'!
dupToStdIn: anSQFileDataStructure
	"Dup a file descriptor to allow it to be attached as the standard input when we
	exec() a new executable. This is Unix specific, in that it assumes that file descriptor
	0 is stdin, 1 is stdout, and 2 is stderr. The dup2() call is used to copy the open file
	descriptors into file descriptors 0, 1 and 2 so that the image which we execute will
	use them as stdin, stdout, and stderr."

	| filenoToDup |
	<returnTypeC: 'void'>
	filenoToDup := self fileDescriptorFrom: anSQFileDataStructure.
	(filenoToDup < 0) ifFalse: [
		(filenoToDup = 0) ifFalse: [
			self cCode: 'fflush(stdin)'.
			self cCode: 'dup2(filenoToDup, 0)'.
			self cCode: 'rewind(stdin)' ]]		"Set stream positioning for stdin."! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl (auto pragmas dtl 2010-11-28) 1/30/2003 22:42'!
dupToStdOut: anSQFileDataStructure
	"Dup a file descriptor to allow it to be attached as the standard output when we
	exec() a new executable. This is Unix specific, in that it assumes that file descriptor
	0 is stdin, 1 is stdout, and 2 is stderr. The dup2() call is used to copy the open file
	descriptors into file descriptors 0, 1 and 2 so that the image which we execute will
	use them as stdin, stdout, and stderr."

	| filenoToDup |
	<returnTypeC: 'void'>
	filenoToDup := self fileDescriptorFrom: anSQFileDataStructure.
	(filenoToDup < 0) ifFalse: [
		(filenoToDup = 1) ifFalse: [
			self cCode: 'fflush(stdout)'.
			self cCode: 'dup2(filenoToDup, 1)' ]]
! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 4/29/2006 11:13'!
fileDescriptorFrom: aSQFileByteArray
	"Answer the OS file descriptor, an integer value, from a SQFile data structure
	byte array, or answer -1 if unable to obtain the file descriptor (probably due
	to receiving an incorrect type of object as aFileHandle)."

	"return type should be int, but skip the declaration to permit inlining"
	(self isSQFileObject: aSQFileByteArray) ifFalse: [^ -1].
	^ self fileno: (self fileHandleFrom: aSQFileByteArray)
! !

!UnixOSProcessPlugin methodsFor: 'primitives - fork and exec' stamp: 'dtl 9/11/2011 09:25'!
forkSqueak
	"Fork a child process, and continue running squeak in the child process.
	Answer the result of the fork() call, either the child pid or zero.

	After calling fork(), two OS processes exist, one of which is the child of the other. On
	systems which implement copy-on-write memory management, and which support the
	fork() system call, both processes will be running Smalltalk images, and will be sharing
	the same memory space. In the original OS process, the resulting value of pid is the
	process id of the child process (a non-zero integer). In the child process, the value of
	pid is zero.

	The child recreates sufficient external resources to continue running. This is done by
	attaching to a new X session. The child is otherwise a copy of the parent process, and
	will continue executing the Smalltalk image at the same point as its parent. The return
	value of this primitive may be used by the two running Smalltalk images to determine
	which is the parent and which is the child.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits.

	The new child image does not start itself from the image in the file system; rather it is
	a clone of the parent image as it existed at the time of primitiveForkSqueak. For this
	reason, the parent and child should agree in advance as to whom is allowed to save the
	image to the file system, otherwise one Smalltalk may overwrite the image of the other.

	This is a simple call to fork(), rather than the more common idiom of vfork() followed
	by exec(). The vfork() call cannot be used here because it is designed to be followed by
	an exec(), and its semantics require the parent process to wait for the child to exit. See
	the BSD programmers documentation for details."

	| pid intervalTimer saveIntervalTimer |
	<export: true>
	<returnTypeC: 'pid_t'>
	<var: 'pid' type: 'pid_t'>
	<var: 'intervalTimer' type: 'struct itimerval'>
	<var: 'saveIntervalTimer' type: 'struct itimerval'>

	"Turn off the interval timer. If this is not done, then the program which we exec in
	the child process will receive a timer interrupt, and will not know how to handle it."
	self cCode: 'intervalTimer.it_interval.tv_sec = 0'.
	self cCode: 'intervalTimer.it_interval.tv_usec = 0'.
	self cCode: 'intervalTimer.it_value.tv_sec = 0'.
	self cCode: 'intervalTimer.it_value.tv_usec = 0'.
	self cCode: 'setitimer (ITIMER_REAL, &intervalTimer, &saveIntervalTimer)'.
	pid := self fork.

	"Enable the timer again before resuming Smalltalk."
	self cCode: 'setitimer (ITIMER_REAL, &saveIntervalTimer, 0L)'.
	^ pid
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 5/9/2011 08:16'!
forwardSignal: sigNum toSemaphoreAt: semaphoreIndex
	"Set a signal handler in the VM which will signal a Smalltalk semaphore at
	semaphoreIndex whenever an external signal sigNum is received. Answer the
	prior value of the signal handler. If semaphoreIndex is zero, the handler is
	unregistered, and the VM returns to its default behavior for handling that
	signal. A handler must be unregistered before it can be registered again.

	The Smalltalk semaphore is expected to be kept at the same index location
	indefinitely during the lifetime of a Squeak session. If that is not the case, the
	handler must be unregistered prior to unregistering the Smalltalk semaphore."

	| oldHandler |
	<returnTypeC: 'void *'>
	<var: 'oldHandler' type: 'void *'>
	semaIndices = nil ifTrue: [^nil]. "Interpreter simulation"
	(semaphoreIndex == 0)
		ifTrue:
			["Disable the handler"
			((semaIndices at: sigNum) ~= 0)
				ifTrue:
					[oldHandler := self originalSignalHandlers at: sigNum.
					oldHandler := self setSignalNumber: sigNum handler: oldHandler.
					semaIndices at: sigNum put: 0.
					^ oldHandler]
				ifFalse: "either -1 for printAllStacks or a positive integer for semaphore forwarding"
					["Signal handler had not been set, answer an error"
					^ self sigErrorNumber]].
	((semaIndices at: sigNum) > 0)
		ifTrue:
			["Handler is already set, answer an error"
			^ self sigErrorNumber].
	oldHandler := self setSignalNumber: sigNum handler: (self handleSignalFunctionAddress).
	(oldHandler ~= self sigErrorNumber)
		ifTrue:
			[self originalSignalHandlers at: sigNum put: oldHandler.
			semaIndices at: sigNum put: semaphoreIndex].
	^ oldHandler
! !

!UnixOSProcessPlugin methodsFor: 'private - support code access' stamp: 'dtl 1/21/2012 09:35'!
getArgumentCount
	"Answer the argument count. If the main module has exported a function called
	getArgCount, call that function. Otherwise try to use a global reference to the
	variable argCnt, which is expected to be declared in the main VM, and may or
	may not be visible depending on platform and compiler conventions. The value
	is cached in a static variable, so this method should not be inlined."

	| argCnt func count |
	<inline: false>
	<var: 'argCnt' type: 'extern int'>
	<var: 'count' declareC: 'static int count = -1'>
	<var: 'func' declareC: 'void (*func)(void)'>
	count < 0 ifTrue: [
		func := interpreterProxy ioLoadFunction: 'ioGetArgCount' From: ''.
		func = nil ifFalse: [count := self cCode: ' ((int (*) (void)) func)()']].
	count < 0 ifTrue: [count := argCnt].
	^count
! !

!UnixOSProcessPlugin methodsFor: 'private - support code access' stamp: 'dtl 1/21/2012 09:25'!
getArgumentVector
	"Answer a pointer to the argument vector. If the main module has exported
	a function called getArgVec, call that function. Otherwise try to use a global
	reference to the variable argVec, which is expected to be declared in the main
	VM, and may or may not be visible depending on platform and compiler
	conventions. The value of the argument vector is cached in a static variable,
	so this method should not be inlined."

	| argVec func ptr |
	<inline: false>
	<returnTypeC: 'char **'>
	<var: 'argVec' type: 'extern char **'>
	<var: 'ptr' declareC: 'static char **ptr = null'>
	<var: 'func' declareC: 'void (*func)(void)'>
	ptr = nil ifTrue: [
		func := interpreterProxy ioLoadFunction: 'ioGetArgVec' From: ''.
		func = nil ifFalse: [ptr := self cCode: ' ((char ** (*) (void)) func)()']].
	ptr = nil ifTrue: [ptr := argVec].
	^ptr
! !

!UnixOSProcessPlugin methodsFor: 'private - support code access' stamp: 'dtl 1/21/2012 09:25'!
getEnvironmentVector
	"Answer a pointer to the environment vector. If the main module has exported
	a function called getEnvVec, call that function. Otherwise try to use a global
	reference to the variable envVec, which is expected to be declared in the main
	VM, and may or may not be visible depending on platform and compiler
	conventions. The value of the environment vector is cached in a static variable,
	so this method should not be inlined."

	| envVec func ptr |
	<inline: false>
	<returnTypeC: 'char **'>
	<var: 'envVec' type: 'extern char **'>
	<var: 'ptr' declareC: 'static char **ptr = null'>
	<var: 'func' declareC: 'void (*func)(void)'>
	ptr = nil ifTrue: [
		func := interpreterProxy ioLoadFunction: 'ioGetEnvVec' From: ''.
		func = nil ifFalse: [ptr := self cCode: ' ((char ** (*) (void)) func)()']].
	ptr = nil ifTrue: [ptr := envVec].
	^ptr
! !

!UnixOSProcessPlugin methodsFor: 'private - support code access' stamp: 'dtl 3/27/2012 22:02'!
getPrintAllStacksFunction
	"Answer the address of the printAllStacks() function. If the main module has
	exported the function address, use that value. Otherwise try to use a global
	reference to the function address, which is expected to be declared in the
	main VM, and may or may not be visible depending on platform and compiler
	conventions. The value of the argument vector is cached in a static variable,
	so this method should not be inlined."

	| func |
	<inline: false>
	<returnTypeC: 'void *'>
	<var: 'func' declareC: 'static void *func'>
	func = nil
		ifTrue: [func := interpreterProxy ioLoadFunction: 'printAllStacks' From: ''.
			func = nil
				ifTrue: [func := self printAllStacksFunctionAddress]].
	^func

! !

!UnixOSProcessPlugin methodsFor: 'simulation support' stamp: 'eem 6/21/2011 15:31'!
getpid
	"Simulate getpid(2)"
	<doNotGenerate>
	^(Smalltalk classNamed: #OSProcessAccessor)
		ifNotNil: [:ospa| ospa  forThisOSProcess primGetPid]
		ifNil: [0]! !

!UnixOSProcessPlugin methodsFor: 'simulation support' stamp: 'eem 6/21/2011 15:31'!
getppid
	"Simulate getppid(2)"
	<doNotGenerate>
	^(Smalltalk classNamed: #OSProcessAccessor)
		ifNotNil: [:ospa| ospa  forThisOSProcess primGetPPid]
		ifNil: [0]! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 5/9/2011 08:11'!
handleSignal: sigNum 
	"This is a signal handler function which runs when a signal is received
	from the operating system. When the signal is received, a Smalltalk Semaphore
	is signaled. This effectively passes the external signal to Squeak to allow
	it to be handled in Smalltalk."

	| semaIndex |
	<returnTypeC: 'void'>
	<var: 'sigNum' type: 'int'>
	semaIndex := semaIndices at: sigNum.
	self forwardSignal: sigNum toSemaphoreAt: semaIndex. "Reset handler"
	self isVmThread
		ifTrue: [semaIndex > 0
				ifTrue: [interpreterProxy signalSemaphoreWithIndex: semaIndex]]
		ifFalse: [self maskForThisThreadAndResend: sigNum]! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/3/2002 18:54'!
handleSignalFunctionAddress

	<returnTypeC: 'void *'>
	^ self cCode: 'handleSignal' inSmalltalk: [#handleSignal:].
! !

!UnixOSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl 12/22/2012 14:29'!
initializeModuleForPlatform
	"Platform specific initialization"

	pidCount := 0.
	self cCode: [self atexit: #sendSignalToPids asSymbol]
		inSmalltalk:
			[semaIndices := CArrayAccessor on: (Array new: 32 withAll: 0).
			originalSigHandlers := CArrayAccessor on: (Array new: 32 withAll: 0)].
	vmThread := self cCode: 'pthread_self()'.
	self isDefined: 'SA_ONSTACK'
		inSmalltalk: []
		comment: 'true if platform supports sigaltstack, else meaningless redeclarations of stack_t and sigaction'
		ifTrue: [self setSigaltstack]
		ifFalse: ["override stack_t declared in sigstack.h because declarations
			cannot easily be removed from setSigaltstack"
			self preprocessorExpression: 'define stack_t sqInt'.
			"likewise provide a compilable but otherwise meaningless redeclaration
			for sigaction in order to permit setSignalNumber:handler: to be compiled on
			platforms that do not support sigaltstack. VirtualMachine was selected
			at random and has no meaning here."
			self preprocessorExpression: 'define sigaction VirtualMachine']! !

!UnixOSProcessPlugin methodsFor: 'private - pthread' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/4/2007 22:28'!
isVmThread
	"Answer true if this method is executing in the context of the pthread in
	which the interpreter executes."

	| thisThread |
	<var: 'thisThread' type: 'pthread_t'>
	thisThread := self cCode: 'pthread_self()' inSmalltalk: [0].
	^ self cCode: 'pthread_equal(thisThread, vmThread)' inSmalltalk: [true]
! !

!UnixOSProcessPlugin methodsFor: 'simulation support' stamp: 'dtl 9/11/2011 17:30'!
k: pid ill: sig
	"Simulate kill(2)"
	<doNotGenerate>
	^0! !

!UnixOSProcessPlugin methodsFor: 'private - pthread' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/4/2006 23:20'!
maskForThisThreadAndResend: sigNum
	"Prevent any future instances of signal sigNum from being delivered
	to this pthread, and reschedule the current signal to be delivered to
	this OS process. Any pthread other than the interpreter thread will
	call this method. Eventually the signal will be delivered to the
	interpreter thread, which will handle it in the context of the interpreter
	thread."

	<var: 'sigNum' type: 'int'>
	self maskSignalForThisThread: sigNum.
	self resendSignal: sigNum! !

!UnixOSProcessPlugin methodsFor: 'private - pthread' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/4/2006 23:20'!
maskSignalForThisThread: sigNum
	"Prevent future invocations of signal sigNum from being delivered to this pthread."

	| sigset |
	<var: 'sigNum' type: 'int'>
	<var: 'sigset' type: 'sigset_t'>
	self cCode: 'sigemptyset(&sigset)'.
	self cCode: 'sigaddset(&sigset, sigNum)'.
	self cCode: 'pthread_sigmask(SIG_BLOCK, &sigset, NULL)'
! !

!UnixOSProcessPlugin methodsFor: 'private - pthread' stamp: 'dtl 3/26/2012 05:43'!
newPthreadTypeByteArray: aPthreadType
	"A pthread_t is an unsigned long, which may be 8 bytes on some 64 bit
	platforms. This stores the value in a byte array that can be easily
	passed to the image."

	| ptr len byteArray |
	<var: 'aPthreadType' type: 'pthread_t'>
	<var: 'ptr' declareC: 'pthread_t *ptr'>
	len := self cCode: 'sizeof(pthread_t)' inSmalltalk: [Smalltalk wordSize].
	byteArray := interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: len.
	ptr := interpreterProxy arrayValueOf: byteArray.
	ptr at: 0 put: aPthreadType.
	^ byteArray
! !

!UnixOSProcessPlugin methodsFor: 'accessing' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/3/2002 19:32'!
originalSignalHandlers
	"An array of signal handler function addresses, one for each signal type. The
	value at each entry is the address of the original signal handler function prior
	to having set a handler."

	<returnTypeC: 'void **'>
	^ originalSigHandlers! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 10/10/2011 08:21'!
primitiveArgumentAt
	"Answer a string containing the OS process argument at index (an Integer) in the
	argument list."

	| index args count sPtr s |
	<export: true>
	<var: 'args' type: 'char **'>
	<var: 'sPtr' type: 'char *'>
	index := interpreterProxy stackIntegerValue: 0.
	count := self getArgumentCount.
	((index > count) or: [index < 1])
	ifTrue:
		[interpreterProxy pop: 2; push: interpreterProxy nilObject]
	ifFalse:
		[args := self getArgumentVector.
		sPtr := self cCode: 'args[index - 1]' inSmalltalk: [args at: index].
		s := self stringFromCString: sPtr.
		interpreterProxy pop: 2; push: s]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:35'!
primitiveCanReceiveSignals: anIntegerPid
	"Send a null signal to the OS process identified by anIntegerPid. Answer false for
	a bad parameter on the stack (the common case is for anIntegerPid equal to nil,
	for which case we should answer false). Answer true if the process exists and can
	receive signals from this process, otherwise false. This test is useful for determining
	if a child process still exists following a Squeak image restart."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>
	(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
		ifTrue: 
			[pidToSignal := interpreterProxy stackIntegerValue: 0.
			result := self cCode: 'kill(pidToSignal, 0)' inSmalltalk: [-1].
			interpreterProxy pop: 2.
			(result == 0)
				ifTrue: [interpreterProxy push: interpreterProxy trueObject]
				ifFalse: [interpreterProxy push: interpreterProxy falseObject]]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 5/3/2012 22:16'!
primitiveChdir
	"Call chdir(2) to change current working directory to the specified path string. Answer
	nil for success, or errno on failure."

	| path errno |
	<export: true>
	<var: 'path' type: 'char *'>
	<var: 'errno' type: 'extern int'>
	path := self transientCStringFromString: (interpreterProxy stackObjectValue: 0).
	(self chdir: path)
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy nilObject]
		ifTrue: [interpreterProxy pop: 2; pushInteger: errno]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - file descriptor' stamp: 'dtl 9/10/2011 14:34'!
primitiveDup: fileDescriptor
	"Call dup() to duplicate a file descriptor using the next available descriptor. Answer
	the new file descriptor or -1 on failure."

	| result fd |
	<export: true>
	<var: 'oldfd' type: 'int'>
	<var: 'result' type: 'int'>
	fd := interpreterProxy stackIntegerValue: 0.
	result := self cCode: 'dup(fd)' inSmalltalk: [-1].
	interpreterProxy pop: 2; pushInteger: result

! !

!UnixOSProcessPlugin methodsFor: 'primitives - file descriptor' stamp: 'dtl (auto pragmas dtl 2010-11-28) 5/17/2009 21:10'!
primitiveDup: oldFileDescriptor To: newFileDescriptor
	"Call dup2() to duplicate a file descriptor. Answer the duplicated file descriptor
	or -1 on failure."

	| result newfd oldfd |
	<export: true>
	<var: 'oldfd' type: 'int'>
	<var: 'newfd' type: 'int'>
	<var: 'result' type: 'int'>
	newfd := interpreterProxy stackIntegerValue: 0.
	oldfd := interpreterProxy stackIntegerValue: 1.
	result := self cCode: 'dup2(oldfd, newfd)' inSmalltalk: [-1].
	interpreterProxy pop: 3; pushInteger: result

! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 10/10/2011 08:20'!
primitiveEnvironmentAt
	"Answer a string containing the OS process environment string at index (an Integer)
	in the environment list."

	| index p envCnt sPtr s |
	<export: true>
	<var: 'p' type: 'char **'>
	<var: 'sPtr' type: 'char *'>
	"Count number of environment variables."
	p := self getEnvironmentVector.
	p = nil ifTrue: [^interpreterProxy primitiveFail].
	envCnt := 0.
	self cCode: 'while (*p++) envCnt++' inSmalltalk: [envCnt := p size].
	p := self getEnvironmentVector. "restore pointer"
	index := interpreterProxy stackIntegerValue: 0.
	((index > envCnt) | (index < 1))
	ifTrue:
		[interpreterProxy pop: 2; push: interpreterProxy nilObject]
	ifFalse:
		[sPtr := self cCode: 'p[index - 1]' inSmalltalk: [p at: index].
		s := self stringFromCString: sPtr.
		interpreterProxy pop: 2; push: s]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 18:58'!
primitiveEnvironmentAtSymbol
	"Answer the value of an environment variable keyed by a Symbol."

	| getenvResult |
	<export: true>
	<var: 'getenvResult' declareC: 'char * getenvResult'>

	getenvResult := self
		getenv: (self transientCStringFromString: (interpreterProxy stackObjectValue: 0)).
	(getenvResult = 0)
		ifTrue:
			[^ interpreterProxy primitiveFail]
		ifFalse:
			[interpreterProxy pop: 2; push: (self stringFromCString: getenvResult)]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 2/21/2002 15:47'!
primitiveErrorMessageAt
	"Answer a string describing an error message"

	| index errMessage p |
	<export: true>
	<var: 'p' declareC: 'char *p'>

	index := interpreterProxy stackIntegerValue: 0.
	p := self cCoerce: (self cCode: 'strerror(index)' inSmalltalk: ['']) to: 'char *'.
	errMessage := self stringFromCString: p.
	interpreterProxy pop: 2; push: errMessage
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:29'!
primitiveFileProtectionMask
	"Call stat(2) to obtain the file protection mask for a file. Answer an Array of
	four integers representing the protection mask, or answer errno on failure. The
	protection mask is four Integers, each of which may be considered an octal digit
	(0-7), with bit values 4, 2, and 1. The first digit selects the set user ID (4) and set
	group ID (2) and save text image (1) attributes. The second digit selects permissions
	for the user who owns the file: read (4), write (2), and execute (1); the third
	selects permissions for other users in the file's group, with the same values; and
	the fourth for other users not in the file's group, with the same values."

	| path statBuf buffer mode errno result |
	<export: true>
	<var: 'path' declareC: 'char *path'>
	<var: 'statBuf' declareC: 'struct stat *statBuf'>
	<var: 'errno' declareC: 'extern int errno'>

	buffer := interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: (self cCode: 'sizeof(struct stat)' inSmalltalk: [1000]).
	statBuf := interpreterProxy arrayValueOf: buffer.
	path := self transientCStringFromString: (interpreterProxy stackObjectValue: 0).
	mode := (self cCode: 'stat(path, statBuf)' inSmalltalk: [0]).
	(mode == 0) ifTrue:
			[mode := self cCode: 'statBuf->st_mode' inSmalltalk: [0].
			result := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 4.
			interpreterProxy
				stObject: result
				at: 4
				put: (interpreterProxy integerObjectOf: (self cCode: 'mode & 07' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 3
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 070) >> 3' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 2
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 0700) >> 6' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 1
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 07000) >> 9' inSmalltalk: [0])).
			interpreterProxy pop: 2; push: result]
		ifFalse:
			[interpreterProxy pop: 2; pushInteger: errno].

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:36'!
primitiveFileStat
	"Call stat(2) to obtain the file protection mask for a file. Answer errno on failure,
	or on success answer an array with: UID with: GID with: protectionMask. The	
	protectionMask is an Array of four integers representing the protection mask, or
	answer errno on failure. The protection mask is four Integers, each of which may
	be considered an octal digit (0-7), with bit values 4, 2, and 1. The first digit selects
	the set user ID (4) and set group ID (2) and save text image (1) attributes. The second
	digit selects permissions for the user who owns the file: read (4), write (2), and
	execute (1); the third selects permissions for other users in the file's group, with
	the same values; and the fourth for other users not in the file's group, with the
	same values."

	| path statBuf buffer uid gid mask mode errno result |
	<export: true>
	<var: 'path' declareC: 'char *path'>
	<var: 'statBuf' declareC: 'struct stat *statBuf'>
	<var: 'errno' declareC: 'extern int errno'>

	result := interpreterProxy
				instantiateClass: interpreterProxy classArray
				indexableSize: 3.
	uid := interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: (self cCode: 'sizeof(uid_t)' inSmalltalk: [Smalltalk wordSize]).
	gid := interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: (self cCode: 'sizeof(gid_t)' inSmalltalk: [Smalltalk wordSize]).
	mask := interpreterProxy
				instantiateClass: interpreterProxy
				classArray indexableSize: 4.
	buffer := interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: (self cCode: 'sizeof(struct stat)' inSmalltalk: [1000]).
	statBuf := interpreterProxy arrayValueOf: buffer.
	path := self transientCStringFromString: (interpreterProxy stackObjectValue: 0).
	mode := (self cCode: 'stat(path, statBuf)' inSmalltalk: [0]).
	(mode == 0) ifTrue:
			[mode := self cCode: 'statBuf->st_mode' inSmalltalk: [0].
			interpreterProxy
				stObject: mask
				at: 4
				put: (interpreterProxy integerObjectOf: (self cCode: 'mode & 07' inSmalltalk: [0])).
			interpreterProxy
				stObject: mask
				at: 3
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 070) >> 3' inSmalltalk: [0])).
			interpreterProxy
				stObject: mask
				at: 2
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 0700) >> 6' inSmalltalk: [0])).
			interpreterProxy
				stObject: mask
				at: 1
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 07000) >> 9' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 3
				put: mask.
			interpreterProxy
				stObject: result
				at: 2
				put: (interpreterProxy integerObjectOf: (self cCode: 'statBuf->st_gid' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 1
				put: (interpreterProxy integerObjectOf: (self cCode: 'statBuf->st_uid' inSmalltalk: [0])).
			interpreterProxy pop: 2; push: result]
		ifFalse:
			[interpreterProxy pop: 2; pushInteger: errno]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - fork and exec' stamp: 'dtl 6/11/2015 20:41'!
primitiveForkExec
	"Fork a child OS process, and do an exec in the child. The parent continues on in
	Smalltalk, and this method answers the pid of the child which was created.

	In this implementation, memory for the argument and environment arrays is allocated
	in the image prior to calling this primitive. This allows us to avoid invoking the
	garbage collector in this primitive (thereby moving the locations of environment
	and argument memory), but comes at the cost of twiddling C pointers here in the
	primitive. An alternative to this whole mess is just to malloc the environment and
	argument vectors, but I think it is a good idea to avoid malloc as much as possible
	so as not to limit future ObjectMemory implementations.

	This primitive replaces #primitiveForkAndExec from earlier versions of the plugin.
	The new name permits backward compatibility for an image running on a VM
	which does not yet have the updated plugin. This implementation uses a different
	argument format on the stack, and differs functionally in that the child now closes
	all file descriptors (including sockets) not required (that is, everything except stdin,
	stdout, and stderr on descriptors 0, 1 and 2). This eliminates some flakey behavior
	in child processes connected to Squeak by pipes, which failed to exit at expected times
	due to the old file descriptors remaining open. This is also cleaner in that garbage
	descriptors are not left hanging around the the child.

	On entry, the stack contains:
		0: workingDir, a null terminated string specifying the working directory to use, or nil.
		1: envOffsets, an array of integers for calculating environment string address offsets.
		2: envVecBuffer, a String buffer containing environment strings arranged to look like char **.
		3: argOffsets, an array of integers for calculating argument string address offsets.
		4: argVecBuffer, a String buffer containing argument strings arranged to look like char **.
		5: stdErr, a ByteArray handle.
		6: stdOut, a ByteArray handle.
		7: stdIn, a ByteArray handle.
		8: executableFile, a null terminated string with the name of the file to execute.
		9: the sender"

	| progNamePtr argsPtr envPtr envVec intervalTimer saveIntervalTimer pid
	envOffsets envVecBuffer argOffsets argVecBuffer stdErr stdOut stdIn executableFile idx
	handleCount envCount argCount envOffsetPtr argOffsetPtr workingDir pwdPtr pwd env args |
	<export: true>
	<var: 'progNamePtr' declareC: 'char *progNamePtr'>
	<var: 'argsPtr' declareC: 'char *argsPtr'>
	<var: 'envPtr' declareC: 'char *envPtr'>
	<var: 'args' declareC: 'char **args'>
	<var: 'env' declareC: 'char **env'>
	<var: 'pwdPtr' declareC: 'char *pwdPtr'>
	<var: 'intervalTimer' type: 'struct itimerval'>
	<var: 'saveIntervalTimer' type: 'struct itimerval'>
	<var: 'envVec' declareC: 'extern char **envVec'>
	<var: 'pid' type: 'pid_t'>
	<var: 'envOffsetPtr' declareC: 'sqInt *envOffsetPtr'>
	<var: 'argOffsetPtr' declareC: 'sqInt *argOffsetPtr'>

	"Do not fork child if running in secure mode"
	(self sandboxSecurity == 1) ifTrue: [interpreterProxy pop: 10; pushInteger: -1] ifFalse: [

	"Turn off the interval timer. If this is not done, then the program which we exec in
	the child process will receive a timer interrupt, and will not know how to handle it."
	self cCode: 'intervalTimer.it_interval.tv_sec = 0'.
	self cCode: 'intervalTimer.it_interval.tv_usec = 0'.
	self cCode: 'intervalTimer.it_value.tv_sec = 0'.
	self cCode: 'intervalTimer.it_value.tv_usec = 0'.
	self cCode: 'setitimer (ITIMER_REAL, &intervalTimer, &saveIntervalTimer)'.

	pid := self fork. "cCode: 'fork()' "
	pid < 0
		ifTrue: [ self perror: 'fork'.
			^self primitiveFail ].
	pid = 0
		ifFalse:
			[ "Normal return to Smalltalk - this is the old parent process."
			"Enable the timer again before resuming Smalltalk."
			self cCode: 'setitimer (ITIMER_REAL, &saveIntervalTimer, 0L)'.
			interpreterProxy pop: 10; pushInteger: pid 
			"Pop 9 arguments plus receiver, push pid."]
		ifTrue:
			[ "This is the new child process"
			workingDir := interpreterProxy stackObjectValue: 0.
			envOffsets := interpreterProxy stackObjectValue: 1.
			envVecBuffer := interpreterProxy stackObjectValue: 2.
			argOffsets := interpreterProxy stackObjectValue: 3.
			argVecBuffer := interpreterProxy stackObjectValue: 4.
			stdErr := interpreterProxy stackObjectValue: 5.
			stdOut := interpreterProxy stackObjectValue: 6.
			stdIn := interpreterProxy stackObjectValue: 7.
			executableFile := interpreterProxy stackObjectValue: 8.

			"If a new working directory has been specified, try to chdir() to it."
			(workingDir ~= interpreterProxy nilObject)
				ifTrue:
					[pwdPtr := interpreterProxy firstIndexableField: workingDir.
					(pwdPtr == 0)
						ifTrue:
							[self cCode: 'fprintf(stderr, "bad workingDir parameter\n")'.
							self cCode: '_exit(-1)']
						ifFalse:
							[pwd := self cCode: 'chdir(pwdPtr)' inSmalltalk: [0].
							(pwd ~= 0)
								ifTrue:
									[self cCode: 'perror("chdir")'.
									self cCode: '_exit(-1)']]].

			"Pointer to name of the file to execute."
			progNamePtr := interpreterProxy arrayValueOf: executableFile.

			"Dup the file handles to attach the new child process to the right streams
			on descriptors 0, 1 and 2."
			(stdErr = interpreterProxy nilObject) ifFalse: [self dupToStdErr: stdErr].
			(stdOut = interpreterProxy nilObject) ifFalse: [self dupToStdOut: stdOut].
			(stdIn = interpreterProxy nilObject) ifFalse: [self dupToStdIn: stdIn].

			"Close all file handles not needed by the child. This is done by brute force
			because the image has no way of being aware of all open file descriptors
			(some are opened by the VM without the knowledge of the image)."
			idx := 3.	"First Unix file descriptor after stdin, stdout, stderr."
			handleCount := self descriptorTableSize.
			[idx < handleCount]
				whileTrue:
					[self cCode: 'close(idx)'.
					idx := idx + 1].

			"Set up the environment, a C array of pointers to C strings, where each string
			is of the form 'KEY=value'. Refer to previously allocated object memory rather
			than using a call to malloc()."
			envVecBuffer = (interpreterProxy nilObject)
				ifTrue:
					[env := envVec]
				ifFalse:
					[envCount := interpreterProxy stSizeOf: envOffsets.
					envPtr := interpreterProxy arrayValueOf: envVecBuffer.
					envOffsetPtr := interpreterProxy firstIndexableField: envOffsets.
					self
						fixPointersInArrayOfStrings: envPtr
						withOffsets: envOffsetPtr
						count: envCount.
					env := self cCoerce: envPtr to: 'char **'].

			"Set up the argument array. Allocate space for a C array of pointers to strings.
			Reer to previously allocated object memory rather than using a call to malloc()."
			argCount := interpreterProxy stSizeOf: argOffsets.
			argsPtr := interpreterProxy arrayValueOf: argVecBuffer.
			argOffsetPtr := interpreterProxy firstIndexableField: argOffsets.
			self
				fixPointersInArrayOfStrings: argsPtr
				withOffsets: argOffsetPtr
				count: argCount.
			args := self cCoerce: argsPtr to: 'char **'.

			"Clean things up before clobbering the running image."
			"Note: If any file descriptors, signal handlers, or other references to external
			resources need to be cleaned up, do it here."
			self restoreDefaultSignalHandlers.

			"Do the exec to overlay the new program on this child address space. This is 
			the end of the world as we know it."
			(self cCode: 'execve(progNamePtr, args, env)' ) = -1
				ifTrue:
					[self cCode: 'perror(progNamePtr)'; cCode: '_exit(-1)' "execve() error, child exits"]
				ifFalse:
					[self cCode: '/* Can''t get here from there */']]]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - fork and exec' stamp: 'dtl 9/11/2011 09:24'!
primitiveForkSqueak
	"Fork a child process, and continue running squeak in the child process. Leave the
	X session connected to the parent process, but close its file descriptor for the child
	process. Open a new X session for the child.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits.
	
	Compatibility - Early versions of OSPP used this primitive. A new implementation
	was provided as primitiveForkSqueakWithoutSigHandler, with the old implementation
	retained as primitiveForkSqueak. The old implementation has now been replaced
	(breaking compatibility for very old versions of OSProcess) such that both primitives
	are identical. Future versions of OSProcess may be changed to refer to primitiveForkSqueak."
	<export: true>
	^self primitiveForkSqueakWithoutSigHandler
! !

!UnixOSProcessPlugin methodsFor: 'primitives - fork and exec' stamp: 'dtl 9/11/2011 09:26'!
primitiveForkSqueakWithoutSigHandler
	"Fork a child process, and continue running squeak in the child process. Leave the
	X session connected to the parent process, but close its file descriptor for the child
	process. Open a new X session for the child.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	| pid |
	<export: true>
	<var: 'pid' type: 'pid_t'>

	"Do not fork child if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 1; pushInteger: -1]
		ifFalse:
			[pid := self forkSqueak.
			interpreterProxy pop: 1; pushInteger: pid]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 2/8/2003 09:51'!
primitiveForwardSignalToSemaphore
	"Set a signal handler in the VM which will signal a Smalltalk semaphore at
	semaphoreIndex whenever an external signal sigNum is received. Answer the
	prior value of the signal handler. If semaphoreIndex is zero or nil, the handler
	is unregistered, and the VM returns to its default behavior for handling that
	signal.

	The Smalltalk semaphore is expected to be kept at the same index location
	indefinitely during the lifetime of a Squeak session. If that is not the case, the
	handler must be unregistered prior to unregistering the Smalltalk semaphore."

	| sigNum index semaphoreIndex handler priorHandler priorHandlerObject hPtr idx |
	<export: true>
	<var: 'handler' declareC: 'void *handler'>
	<var: 'priorHandler' declareC: 'union {void *handler; unsigned char bytes[sizeof(void *)];} priorHandler'>
	<var: 'hPtr' declareC: 'char *hPtr'>
	index := interpreterProxy stackValue: 0.
	(index == interpreterProxy nilObject)
		ifTrue:
			[semaphoreIndex := 0]
		ifFalse:
			[(interpreterProxy isIntegerObject: index)
				ifTrue:
					[semaphoreIndex := interpreterProxy stackIntegerValue: 0]
				ifFalse:
					[^ interpreterProxy primitiveFail]].
	sigNum := interpreterProxy stackIntegerValue: 1.
	handler := self forwardSignal: sigNum toSemaphoreAt: semaphoreIndex.
	(handler == self sigErrorNumber) ifTrue: [^ interpreterProxy primitiveFail].
	priorHandlerObject := interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: self sizeOfPointer.
	hPtr := interpreterProxy arrayValueOf: priorHandlerObject.
	self cCode: 'priorHandler.handler = handler'.
	idx := 0.
	[idx < self sizeOfPointer] whileTrue:
		[self cCode: 'hPtr[idx] = priorHandler.bytes[idx]'.
		idx := idx + 1].
	interpreterProxy pop: 3 thenPush: priorHandlerObject

! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 19:07'!
primitiveGetCurrentWorkingDirectory
	"Answer a string containing the current working directory."

	| cwd cwdString bufferSize incrementBy maxSize buffer |
	<export: true>
	<var: 'buffer' declareC: 'char *buffer'>
	<var: 'cwd' declareC: 'char *cwd'>

	bufferSize := 100.
	incrementBy := 100.
	maxSize := 5000.

	[cwdString := interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: bufferSize.
	buffer := interpreterProxy arrayValueOf: cwdString.
	cwd := self get: buffer cwd: bufferSize. "getcwd(buffer, bufferSize)"
	((cwd == 0)
		and: [bufferSize < maxSize])]
			whileTrue:
				[bufferSize := bufferSize + incrementBy].
	(cwd == 0)
		ifTrue:
			[^ interpreterProxy primitiveFail]
		ifFalse:
			[cwdString := self stringFromCString: cwd.
			interpreterProxy pop: 1; push: cwdString]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/18/2007 10:14'!
primitiveGetEGid
	"Answer the effective group ID of my OS process"

	| eGid |
	<export: true>
	<var: 'eGid' type: 'gid_t'>
	eGid := self getegid.
	interpreterProxy pop: 1; pushInteger: eGid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/18/2007 10:16'!
primitiveGetEUid
	"Answer the effective user ID of my OS process"

	| eUid |
	<export: true>
	<var: 'eUid' type: 'uid_t'>
	eUid := self geteuid.
	interpreterProxy pop: 1; pushInteger: eUid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/18/2007 10:16'!
primitiveGetGid
	"Answer the group ID of my OS process"

	| gid |
	<export: true>
	<var: 'gid' type: 'gid_t'>
	gid := self getgid.
	interpreterProxy pop: 1; pushInteger: gid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/15/2007 19:45'!
primitiveGetPGid
	"Answer the process group ID of the process identified by pid"

	| pid pgid |
	<export: true>
	<var: 'pid' type: 'pid_t'>
	<var: 'pgid' type: 'pid_t'>
	pid := interpreterProxy stackIntegerValue: 0.
	pgid := self getpgid: pid.
	pgid == -1 ifTrue: [^ interpreterProxy primitiveFail].
	interpreterProxy pop: 2; pushInteger: pgid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/18/2007 10:18'!
primitiveGetPGrp
	"Answer the process group ID of this OS process"

	| pgid |
	<export: true>
	<var: 'pgid' type: 'pid_t'>
	pgid := self getpgrp.
	pgid == -1 ifTrue: [^ interpreterProxy primitiveFail].
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: pgid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 5/8/2011 22:14'!
primitiveGetPPid
	"Answer the process ID of the parent process of my OS process"

	| ppid |
	<export: true>
	<var: 'ppid' type: 'pid_t'>
	ppid := self cCode: 'getppid()' inSmalltalk: [0].
	interpreterProxy pop: 1; pushInteger: ppid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 5/8/2011 22:04'!
primitiveGetPid
	"Answer the process ID of my OS process"

	| pid |
	<export: true>
	<var: 'pid' type: 'pid_t'>
	pid := self cCode: 'getpid()' inSmalltalk: [0].
	interpreterProxy pop: 1; pushInteger: pid! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 6/13/2005 00:23'!
primitiveGetStdErrHandle
	"Answer the file handle for standard error of my OS process"

	| file thisSession fileOop |
	<export: true>
	<var: 'file' declareC: 'SQFile *file'>
	<var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'>
	fileOop := self newSQFileByteArray.
	file := self fileValueOf: fileOop.
	thisSession := self getThisSessionIdentifier.
	self cCode: 'file->file = stderr'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy pop: 1; push: fileOop
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 6/13/2005 00:23'!
primitiveGetStdInHandle
	"Answer the file handle for standard input of my OS process"

	| file thisSession fileOop |
	<export: true>
	<var: 'file' declareC: 'SQFile *file'>
	<var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'>
	fileOop := self newSQFileByteArray.
	file := self fileValueOf: fileOop.
	thisSession := self getThisSessionIdentifier.
	self cCode: 'file->file = stdin'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 0'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy pop: 1; push: fileOop
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 6/13/2005 00:23'!
primitiveGetStdOutHandle
	"Answer the file handle for standard output of my OS process"

	| file thisSession fileOop |
	<export: true>
	<var: 'file' declareC: 'SQFile *file'>
	<var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'>
	fileOop := self newSQFileByteArray.
	file := self fileValueOf: fileOop.
	thisSession := self getThisSessionIdentifier.
	self cCode: 'file->file = stdout'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy pop: 1; push: fileOop
! !

!UnixOSProcessPlugin methodsFor: 'primitives - pthreads' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/5/2006 10:50'!
primitiveGetThreadID
	"Answer the ID of the pthread that is currently executing (the interpreter thread).
	A thread ID may be a 64 bit value on some platforms, so answer a byte array
	containing the value in machine-dependent byte order."

	<export: true>
	interpreterProxy pop: 1 thenPush: (self newPthreadTypeByteArray: vmThread)
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/18/2007 10:19'!
primitiveGetUid
	"Answer the user ID of my OS process"

	| uid |
	<export: true>
	<var: 'uid' type: 'uid_t'>
	uid := self getuid.
	interpreterProxy pop: 1; pushInteger: uid! !

!UnixOSProcessPlugin methodsFor: 'primitives - VM atexit' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/18/2007 10:34'!
primitiveKillOnExit
	"Set a list of pids to kill with signum when VM exits. If the signum
	parameter is nil, the default value of SIGTERM will be used."

	| pids signum pidPointer count p pid |
	<export: true>
	<var: 'pidPointer' type: 'sqInt *'>
	<var: 'p' type: 'pid_t *'>
	<var: 'pid' type: 'pid_t'>
	pids := interpreterProxy stackValue: 1.
	signum := interpreterProxy stackValue: 0.
	pidCount := interpreterProxy stSizeOf: pids.
	interpreterProxy failed ifTrue: [^ nil].
	pidArray == 0 ifFalse: [self free: pidArray].
	pidArray := self cCode: 'malloc(pidCount * sizeof(pid_t))'.
	pidArray == 0
		ifTrue: [pidCount := 0.
			^ interpreterProxy primitiveFail].
	"Copy pid values into externally allocated array"
	pidPointer := interpreterProxy firstIndexableField: pids.
	count := 0.
	p := pidArray.
	[count <= pidCount]
		whileTrue: [pid := interpreterProxy
						integerValueOf: (self cCode: '*pidPointer++').
			self cCode: '*p++ = pid'.
			count := count + 1].
	signum == interpreterProxy nilObject
		ifFalse: [sigNumToSend := interpreterProxy checkedIntegerValueOf: signum].
	interpreterProxy failed
		ifTrue: [pidCount := 0].
	interpreterProxy pop: 2! !

!UnixOSProcessPlugin methodsFor: 'primitives - file locking' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:31'!
primitiveLockFileRegion
	"Take a struct SQFile from the stack, and request a lock on the specified region.
	If the exclusive flag is true, then request an exclusive (F_WRLCK) lock on the
     file. Otherwise, request a shared (F_RDLCK) lock. Any number of Unix processes
     may hold  a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock). Answer the result of the call to fcntl().

	If length is zero, then the entire file will be locked, including region extents that
	have not yet been allocated for the file."

	| lockStruct exclusive len start sqFileOop fileNo result |
	<export: true>
	<var: 'lockStruct' declareC: 'struct flock lockStruct'>
	<var: 'fileNo' type: 'int'>
	<var: 'result' type: 'int'>

	"Get the parameters from the stack"
	exclusive := (interpreterProxy stackValue: 0) == interpreterProxy trueObject.
	len := interpreterProxy stackIntegerValue: 1.
	start := interpreterProxy stackIntegerValue: 2.
	sqFileOop := interpreterProxy stackValue: 3.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	fileNo := self unixFileNumber: (self fileHandleFrom: sqFileOop).

	"Set up the flock structure parameter for fcntl()"
	exclusive
		ifTrue: [self cCode: 'lockStruct.l_type = F_WRLCK']
		ifFalse: [self cCode: 'lockStruct.l_type = F_RDLCK'].
	self cCode: 'lockStruct.l_whence = SEEK_SET'.
	self cCode: 'lockStruct.l_start = start'.
	self cCode: 'lockStruct.l_len = len'.
	self cCode: 'lockStruct.l_pid = 0'.

	"Request the lock and answer the result of the fcntl call"
	result := self cCode: 'fcntl(fileNo, F_SETLK, &lockStruct)'.
	interpreterProxy pop: 5; pushInteger: result
! !

!UnixOSProcessPlugin methodsFor: 'primitives - mktemp' stamp: 'dtl 6/3/2013 19:06'!
primitiveMkdtemp
	"Call mkdtemp() to generate a uniquely named temporary directory from a template
	string. The last six characters of template must be XXXXXX and these are replaced
	with characters that make the directory name unique. The directory is then created
	with permissions 0700. Answer the new directory name."

	| sourceString templateString template strPtr |
	<export: true>
	<var: 'template' type: 'char *'>
	<var: 'strPtr' type: 'char *'>

	sourceString := interpreterProxy stackObjectValue: 0.
	(interpreterProxy isWordsOrBytes: sourceString)
		ifFalse: [^self primitiveFailFor: PrimErrBadArgument].
	templateString := self nullTerminatedStringFromString: sourceString.
	template := interpreterProxy arrayValueOf: templateString.
	strPtr := self mkdtemp: template.
	strPtr = 0
		ifTrue: [self perror: 'mkdtemp'.
			^interpreterProxy primitiveFail]
		ifFalse: [interpreterProxy pop: 2 thenPush: templateString]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - mktemp' stamp: 'dtl 6/3/2013 22:57'!
primitiveMkstemp
	"Call mkstemp() to generate a unique temporary file from a template string.
	The last six characters of template must be XXXXXX and these are replaced with
	a string that make the filename unique. Answer an SQFile descriptor for the newly
	opened file.
	
	Contents of the template string are modified, and provide the name of the newly
	created file. The file is created with permissions 0600, that is, read plus write
	for owner only."

	| sourceString templateString template fileDescriptor |
	<export: true>
	<var: 'template' type: 'char *'>
	<var: 'handle' type: 'FILEHANDLETYPE'>
	<var: 'file' declareC: 'SQFile *file'>
	sourceString := interpreterProxy stackObjectValue: 0.
	(interpreterProxy isWordsOrBytes: sourceString)
		ifFalse: [^self primitiveFailFor: PrimErrBadArgument].
	templateString := self nullTerminatedStringFromString: sourceString.
	template := interpreterProxy arrayValueOf: templateString.
	fileDescriptor := self mkstemp: template.
	fileDescriptor < 0
		ifTrue: [self perror: 'mkstemp'.
			^interpreterProxy primitiveFail]
		ifFalse: [ | fileOop file thisSession handle |
			fileOop := self newSQFileByteArray. "A SQFile data structure for the file descriptor"
			file := self fileValueOf: fileOop.
			thisSession := self getThisSessionIdentifier.
			handle := self fd: fileDescriptor open: 'r+'.
			self cCode: 'file->file = handle'.
			self cCode: 'file->sessionID = thisSession'.
			self cCode: 'file->writable = 1'.
			self cCode: 'file->lastOp = 0'.
			interpreterProxy pop: 2 thenPush: fileOop]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - mktemp' stamp: 'dtl 6/3/2013 19:01'!
primitiveMktemp
	"Call mktemp() to generate a unique temporary filename from a template string.
	The last six characters of template must be XXXXXX and these are replaced with
	a string that make the filename unique.
	
	Use primitiveMkstemp to avoid file creation race condition, see man 3 mktemp."

	| sourceString templateString template strPtr |
	<export: true>
	<var: 'template' type: 'char *'>
	<var: 'strPtr' type: 'char *'>

	sourceString := interpreterProxy stackObjectValue: 0.
	(interpreterProxy isWordsOrBytes: sourceString)
		ifFalse: [^self primitiveFailFor: PrimErrBadArgument].
	templateString := self nullTerminatedStringFromString: sourceString.
	template := interpreterProxy arrayValueOf: templateString.
	strPtr := self mktemp: template.
	(strPtr at: 0) = 0
		ifTrue: [ "pointer to empty string means mktemp failed"
			^interpreterProxy primitiveFail]
		ifFalse: [interpreterProxy pop: 2 thenPush: templateString]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 10/8/2005 09:43'!
primitiveNice
	"Change the scheduling priority of this process by the given nice increment.
	A positive increment decreases the priority. Only the superuser can specify
	a negative value (to increase the priority). See man(2) nice.

	Different versions of Unix are inconsistent in their return values. The only
	reliable test for success is to clear errno prior to the call, and test its value
	if the result of nice() is -1."

	| niceIncrement errno result |
	<export: true>
	<var: 'errno' type: 'extern int'>
	<var: 'niceIncrement' type: 'int'>
	niceIncrement := interpreterProxy stackIntegerValue: 0.
	errno := 0.
	result := self nice: niceIncrement.
	(result == -1) ifTrue:
		["sys call may have failed, test errno to be sure"
		(errno == 0) ifFalse:
			[^ interpreterProxy primitiveFail]].
	interpreterProxy pop: 2; pushInteger: result.
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 7/19/2010 22:00'!
primitivePrintAllStacksOnSignal
	"Set a signal handler in the VM which will print all call stacks on the console
	output whenever an external signal sigNum is received. Answer the
	prior value of the signal handler."

	| sigNum handler priorHandler priorHandlerObject hPtr idx |
	<export: true>
	<var: 'handler' declareC: 'void *handler'>
	<var: 'priorHandler' declareC: 'union {void *handler; unsigned char bytes[sizeof(void *)];} priorHandler'>
	<var: 'hPtr' declareC: 'char *hPtr'>
	sigNum := interpreterProxy stackIntegerValue: 0.
	handler := self printAllStacksOnSignal: sigNum.
	(handler == self sigErrorNumber) ifTrue: [^ interpreterProxy primitiveFail].
	priorHandlerObject := interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: self sizeOfPointer.
	hPtr := interpreterProxy arrayValueOf: priorHandlerObject.
	self cCode: 'priorHandler.handler = handler'.
	idx := 0.
	[idx < self sizeOfPointer] whileTrue:
		[self cCode: 'hPtr[idx] = priorHandler.bytes[idx]'.
		idx := idx + 1].
	interpreterProxy pop: 2 thenPush: priorHandlerObject

! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/18/2007 10:20'!
primitivePutEnv
	"Set an environment variable using a string of the form 'KEY=value'. This
	implementation allocates a C string using malloc to allocate from the C heap
	(using cStringFromString rather than transientCStringFromString). This
	is necessary because the C runtime library does not make a copy of the
	string into separately allocated environment memory."

	| cStringPtr keyValueString |
	<export: true>
	<var: 'cStringPtr' declareC: 'char *cStringPtr'>
	keyValueString := interpreterProxy stackObjectValue: 0.
	cStringPtr := self cStringFromString: keyValueString.
	((self putenv: cStringPtr) == 0)	"Set environment variable."
		ifTrue: [interpreterProxy pop: 2; push: keyValueString]
		ifFalse: [^ interpreterProxy primitiveFail]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/3/2003 15:36'!
primitiveRealpath
	"Answer the real path for a path string as determined by realpath()."

	|  bufferSize pathString newPathString buffer realpathResult s |
	<export: true>
	<var: 'pathString' declareC: 'char *pathString'>
	<var: 'buffer' declareC: 'char *buffer'>
	<var: 'realpathResult' declareC: 'char * realpathResult'>

	bufferSize := 1024.
	newPathString := interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: bufferSize.
	interpreterProxy pushRemappableOop: newPathString.
	pathString := self transientCStringFromString: (interpreterProxy stackObjectValue: 0).
	newPathString := interpreterProxy popRemappableOop.
	buffer := interpreterProxy arrayValueOf: newPathString.
	realpathResult := self cCode: 'realpath(pathString, buffer)' inSmalltalk: [''].
	(realpathResult = 0)
		ifTrue:
			[^ interpreterProxy primitiveFail]
		ifFalse:
			[((self strlen: realpathResult) >= 1024) ifTrue:
				[self perror: 'warning: statically allocated array exceeded in UnixOSProcessPlugin>>primitiveRealPath, object memory may have been corrupted'.
				^ interpreterProxy primitiveFail].
			s := self stringFromCString: realpathResult.
			interpreterProxy pop: 2; push: s]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 1/3/2011 20:44'!
primitiveReapChildProcess
	"Clean up after the death of a child, and answer an array with the pid and
	the exit status of the child process. Answer nil if the pidToHandle does not exist."

	| pidToHandle pidResult exitStatus resultArray arrayPtr pid status |
	<export: true>
	<var: 'exitStatus' type: 'int'>
	<var: 'pidResult' type: 'pid_t'>
	<var: 'arrayPtr' declareC: 'sqInt *arrayPtr'>
	exitStatus := 0.	"Force C code translator to declare the variable"
	pidToHandle := interpreterProxy stackIntegerValue: 0.
	pidResult := self cCode: 'waitpid ( pidToHandle, &exitStatus, WNOHANG )'
					inSmalltalk: [ exitStatus := -1 ].
	pidResult <= 0
		ifTrue: [ interpreterProxy pop: 2; push: interpreterProxy nilObject ]
		ifFalse: [ "Answer an array with pid and result status "
			pid := interpreterProxy integerObjectOf: pidResult.
			status := interpreterProxy integerObjectOf: exitStatus.
			resultArray := interpreterProxy
							instantiateClass: interpreterProxy classArray
							indexableSize: 2.
			arrayPtr := interpreterProxy firstIndexableField: resultArray.
			self cCode: 'arrayPtr[0] = pid'
				inSmalltalk: [resultArray at: 1 put: pidResult].
			self cCode: 'arrayPtr[1] = status'
				inSmalltalk: [resultArray at: 2 put: exitStatus].
			interpreterProxy pop: 2; push: resultArray ]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/27/2006 09:33'!
primitiveSQFileFlush
	"Take a struct SQFile from the stack, and call fflush() to flush the OS stream. This flushes the
	file stream in the C library, not the stream in Smalltalk. For output streams, consider setting
	the OS stream (C library) to unbuffered output, and letting Smalltalk do all the buffering."

	| sqFileOop result |
	<export: true>
	<var: 'result' type: 'int'>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	result := self fflush: (self fileHandleFrom: sqFileOop).
	interpreterProxy pop: 2; pushInteger: result
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:28'!
primitiveSQFileSetBlocking
	"Take a struct SQFile from the stack, and call fcntl() to set the file for blocking I/O."

	| sqFileOop descriptor flags retVal |
	<export: true>
	<var: 'flags' type: 'int'>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	descriptor := self fileDescriptorFrom: sqFileOop.
	(descriptor = -1) ifTrue: [^ interpreterProxy primitiveFail].
	flags := self cCode: 'fcntl(descriptor, F_GETFL)'.
	retVal := self cCode: 'fcntl(descriptor, F_SETFL, flags & ~O_NONBLOCK)'.
	interpreterProxy pop: 2; pushInteger: retVal

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 5/23/2009 11:58'!
primitiveSQFileSetNonBlocking
	"Take a struct SQFile from the stack, and call fcntl() to set the file non-blocking I/O."

	| sqFileOop descriptor flags retVal |
	<export: true>
	<var: 'flags' type: 'int'>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	descriptor := self fileDescriptorFrom: sqFileOop.
	(descriptor < 0) ifTrue: [^ interpreterProxy primitiveFail].
	flags := self cCode: 'fcntl(descriptor, F_GETFL)'.
	retVal := self cCode: 'fcntl(descriptor, F_SETFL, flags | O_NONBLOCK)'.
	interpreterProxy pop: 2; pushInteger: retVal

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/27/2006 09:45'!
primitiveSQFileSetUnbuffered
	"Take a struct SQFile from the stack, and call setbuf() to set the OS file stream
	(implemented in the C library) for unbuffered I/O. Answers the result of a fflush()
	call, not the result of the setbuf() call (which is type void). This is nearly useless,
	but may at least provide an indicator that we are operating on a valid file stream."

	| sqFileOop retVal file |
	<export: true>
	<var: 'file' type: 'FILEHANDLETYPE'>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	file := self fileHandleFrom: sqFileOop.
	retVal := self fflush: file.
	self set: file buf: 0. "sebuf(file, NULL)"
	interpreterProxy pop: 2; pushInteger: retVal

! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 5/9/2011 08:11'!
primitiveSemaIndexFor
	"Answer the registration index of the semaphore currently associated with the
	signal handler for sigNum."

	| sigNum index |
	<export: true>
	sigNum := interpreterProxy stackIntegerValue: 0.
	semaIndices = nil
		ifTrue: [index := 0 "interpreter simulation"]
		ifFalse: [index := semaIndices at: sigNum].
	interpreterProxy pop: 2; pushInteger: index
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 17:37'!
primitiveSendSigabrtTo: anIntegerPid 
	"Send SIGABRT (abort) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigAbrtNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:09'!
primitiveSendSigalrmTo: anIntegerPid 
	"Send SIGALRM (alarm clock) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigAlrmNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:10'!
primitiveSendSigchldTo: anIntegerPid
	"Send SIGCHLD (child status has changed, usually death of child) to the OS process
	identified by anIntegerPid. Use an explicit check for isIntegerObject so we can
	return -1 on error (the stackIntegerValue: method answers 1 on error, and 1 is a
	valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigChldNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:11'!
primitiveSendSigcontTo: anIntegerPid
	"Send SIGCONT (continue) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigContNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:12'!
primitiveSendSighupTo: anIntegerPid
	"Send SIGHUP (hangup) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigHupNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:12'!
primitiveSendSigintTo: anIntegerPid
	"Send SIGINT (interrupt) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigIntNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:12'!
primitiveSendSigkillTo: anIntegerPid
	"Send SIGKILL (kill, unblockable) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigKillNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:12'!
primitiveSendSigpipeTo: anIntegerPid
	"Send SIGPIPE (broken pipe) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigPipeNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:13'!
primitiveSendSigquitTo: anIntegerPid
	"Send SIGQUIT (quit) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigQuitNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:14'!
primitiveSendSigstopTo: anIntegerPid
	"Send SIGSTOP (stop, unblockable) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigStopNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:14'!
primitiveSendSigtermTo: anIntegerPid
	"Send SIGTERM (termination) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigTermNumber toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:15'!
primitiveSendSigusr1To: anIntegerPid
	"Send SIGUSR1 (User-defined signal 1) to the OS process identified by anIntegerPid. Use
	an explicit check for isIntegerObject so we can return -1 on error (the stackIntegerValue:
	method answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigUsr1Number toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/17/2007 18:15'!
primitiveSendSigusr2To: anIntegerPid
	"Send SIGUSR2 (User-defined signal 2) to the OS process identified by anIntegerPid. Use
	an explicit check for isIntegerObject so we can return -1 on error (the stackIntegerValue:
	method answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	<export: true>
	<var: 'pidToSignal' type: 'pid_t'>
	<var: 'result' type: 'int'>

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal := interpreterProxy stackIntegerValue: 0.
					result := self sendSignal: self sigUsr2Number toPid: pidToSignal.
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/15/2007 19:36'!
primitiveSetPGid
	"Set the process group ID of the process identified by pid to a new process
	group ID."

	| pid pgid |
	<export: true>
	<var: 'pid' type: 'pid_t'>
	<var: 'pgid' type: 'pid_t'>
	pid := interpreterProxy stackIntegerValue: 1.
	pgid := interpreterProxy stackIntegerValue: 0.
	(self set: pid pgid: pgid) == -1 ifTrue: [^ interpreterProxy primitiveFail].
	interpreterProxy pop: 2
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/20/2007 19:04'!
primitiveSetPGrp
	"Set a new process group for this OS process. Newly created child processes
	will be members of the new process group. Note: Use setpgid(0,0) rather than
	the equivalent setpgrp() because setpgrp() is implemented differently on
	some flavors of Unix."

	<export: true>
	(self set: 0 pgid: 0) == -1 ifTrue: [^ interpreterProxy primitiveFail]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/18/2007 10:12'!
primitiveSetSid
	"Quoted from Linux man pages:
	setsid() creates a new session if the calling process is not a process group leader.
	The calling process is the leader of the new session, the process  group  leader  of
	the new process group, and has no controlling tty.  The process group ID and session
	ID of the calling process are set to the PID of the calling  process.   The  calling
	process  will be the only process in this new process group and in this new session."

	| sessionId |
	<export: true>
	<var: 'sessionId' type: 'pid_t'>
	sessionId := self setsid.
	sessionId == -1 ifTrue: [^ interpreterProxy primitiveFail].
	interpreterProxy pop: 1; pushInteger: sessionId
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/4/2002 00:26'!
primitiveSigChldNumber
	"Integer value corresponding to SIGCHLD"

	<export: true>
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigChldNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 12/6/2002 17:56'!
primitiveSigHupNumber
	"Integer value corresponding to SIGHUP"

	<export: true>
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigHupNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/4/2002 00:26'!
primitiveSigIntNumber
	"Integer value corresponding to SIGINT"

	<export: true>
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigIntNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 12/6/2002 17:57'!
primitiveSigKillNumber
	"Integer value corresponding to SIGKILL"

	<export: true>
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigKillNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/4/2002 00:27'!
primitiveSigPipeNumber
	"Integer value corresponding to SIGPIPE"

	<export: true>
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigPipeNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 12/6/2002 17:57'!
primitiveSigQuitNumber
	"Integer value corresponding to SIGQUIT"

	<export: true>
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigQuitNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 12/6/2002 17:58'!
primitiveSigTermNumber
	"Integer value corresponding to SIGTERM"

	<export: true>
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigTermNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/4/2005 06:39'!
primitiveSigUsr1Number
	"Integer value corresponding to SIGUSR1"

	<export: true>
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigUsr1Number
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/4/2005 06:40'!
primitiveSigUsr2Number
	"Integer value corresponding to SIGUSR2"

	<export: true>
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigUsr2Number
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 12/31/2010 10:05'!
primitiveTestEndOfFileFlag
	"Take a struct SQFile from the stack, and call feof(3) to determine if the file has
	reached end of file. The flag is set only by a previous read operation, so end of
	file is not detected until an actual EOF condition has been detected by a read attempt."

	| sqFileOop file result |
	<export: true>
	<var: 'file' type: 'FILEHANDLETYPE'>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	file := self fileHandleFrom: sqFileOop.
	(file = 0) ifTrue: [^ interpreterProxy primitiveFail].
	(self feof: file)
		ifTrue: [result := interpreterProxy trueObject]
		ifFalse: [result := interpreterProxy falseObject].
	interpreterProxy pop: 2; push: result

! !

!UnixOSProcessPlugin methodsFor: 'primitives - file locking' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:34'!
primitiveTestLockableFileRegion
	"Take a struct SQFile from the stack, and check for ability to lock the specified region.
	If the exclusive flag is true, then specify an exclusive (F_WRLCK) lock on the
     file. Otherwise, specify a shared (F_RDLCK) lock. Any number of Unix processes
     may hold  a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock).

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	If the fcntl() call fails, answer -1 (the result of the failed call). Otherwise,
	answer an array with the following six fields:
		lockable (true or false)
		l_pid (pid of the process preventing this lock request, or nil)
		l_type (request type F_WRLCK or F_RDLOCK of the process preventing this lock request)
		l_whence (the SEEK_SET, SEEK_CUR, or SEEK_END value of the lock preventing this lock request).
		l_start (offset of the region lock preventing this lock request)
		l_len (length of the region lock preventing this lock request)"

	| lockStruct exclusive len start sqFileOop fileNo result resultArray canObtainLock |
	<export: true>
	<var: 'lockStruct' declareC: 'struct flock lockStruct'>
	<var: 'fileNo' type: 'int'>
	<var: 'result' type: 'int'>

	"Get the parameters from the stack"
	exclusive := (interpreterProxy stackValue: 0) == interpreterProxy trueObject.
	len := interpreterProxy stackIntegerValue: 1.
	start := interpreterProxy stackIntegerValue: 2.
	sqFileOop := interpreterProxy stackValue: 3.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	fileNo := self unixFileNumber: (self fileHandleFrom: sqFileOop).

	"Set up the flock structure parameter for fcntl()"
	exclusive
		ifTrue: [self cCode: 'lockStruct.l_type = F_WRLCK']
		ifFalse: [self cCode: 'lockStruct.l_type = F_RDLCK'].
	self cCode: 'lockStruct.l_whence = SEEK_SET'.
	self cCode: 'lockStruct.l_start = start'.
	self cCode: 'lockStruct.l_len = len'.
	self cCode: 'lockStruct.l_pid = 0'.

	"Check availability of the lock"
	result := self cCode: 'fcntl(fileNo, F_GETLK, &lockStruct)'.
	(result == -1)
		ifTrue:
			[interpreterProxy pop: 5; pushInteger: result]
		ifFalse:
			[(self cCode: 'lockStruct.l_type == F_UNLCK')
				ifTrue: [canObtainLock := interpreterProxy trueObject]
				ifFalse: [canObtainLock := interpreterProxy falseObject].
			
resultArray := interpreterProxy
							instantiateClass: (interpreterProxy classArray)
							indexableSize: 6.
			interpreterProxy
				stObject: resultArray
				at: 1
				put: canObtainLock.
			interpreterProxy
				stObject: resultArray
				at: 2
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_pid')).
			interpreterProxy
				stObject: resultArray
				at: 3
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_type')).
			interpreterProxy
				stObject: resultArray
				at: 4
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_whence')).
			interpreterProxy
				stObject: resultArray
				at: 5
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_start')).
			interpreterProxy
				stObject: resultArray
				at: 6
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_len')).
			interpreterProxy pop: 5 thenPush: resultArray]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - file descriptor' stamp: 'dtl (auto pragmas dtl 2010-11-28) 5/17/2009 20:39'!
primitiveUnixFileClose: anIntegerFileNumber
	"Close a file handle at the close(2) level, using a handle returned by
	#primitiveUnixFileNumber."

	|  handle result |
	<export: true>
	<var: 'handle' type: 'int'>
	<var: 'result' type: 'int'>
	handle := interpreterProxy stackIntegerValue: 0.
	result := self cCode: 'close(handle)' inSmalltalk: [-1].
	interpreterProxy pop: 2; pushInteger: result

! !

!UnixOSProcessPlugin methodsFor: 'primitives - file descriptor' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/27/2006 07:49'!
primitiveUnixFileNumber
	"Take a struct SQFile from the stack, and answer the value of its Unix file number.
	The Unix file number is not directly useful to Squeak, but may be interesting for
	debugging problems involving failure to close unused file handles."

	| sqFileOop fileNo |
	<export: true>
	<var: 'fileNo' type: 'int'>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	fileNo := self unixFileNumber: (self fileHandleFrom: sqFileOop).
	interpreterProxy pop: 2; pushInteger: fileNo
! !

!UnixOSProcessPlugin methodsFor: 'primitives - file locking' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:37'!
primitiveUnlockFileRegion
	"Take a struct SQFile from the stack, and unlock the specified region.
	Answer the result of the call to fcntl(). If the region is in the file lock cache,
	remove it, but otherwise ignore the cache. The cache supports Win32 semantics
	within a single Squeak image, but not across separate images, therefore the
	unlock should be attempted regardless of whether this image thinks that the
	region has previously been locked. Answer the result of the call to fcntl()."

	| lockStruct len start sqFileOop fileNo result |
	<export: true>
	<var: 'lockStruct' declareC: 'struct flock lockStruct'>
	<var: 'fileNo' type: 'int'>
	<var: 'result' type: 'int'>

	"Get the parameters from the stack"
	len := interpreterProxy stackIntegerValue: 0.
	start := interpreterProxy stackIntegerValue: 1.
	sqFileOop := interpreterProxy stackValue: 2.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	fileNo := self unixFileNumber: (self fileHandleFrom: sqFileOop).

	"Set up the flock structure parameter for fcntl()"
	self cCode: 'lockStruct.l_type = F_UNLCK'.
	self cCode: 'lockStruct.l_whence = SEEK_SET'.
	self cCode: 'lockStruct.l_start = start'.
	self cCode: 'lockStruct.l_len = len'.
	self cCode: 'lockStruct.l_pid = 0'.

	"Request the lock and answer the result of the fcntl call"
	result := self cCode: 'fcntl(fileNo, F_SETLK, &lockStruct)'.
	interpreterProxy pop: 4; pushInteger: result
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 17:36'!
primitiveUnsetEnv
	"Unset an environment variable."

	"FIXME: unsetenv() is not portable. For Solaris or any other system which does not
	support unsetenv(), just comment it out in this method and rebuild the plugin."

	<export: true>
	self unsetenv: (self transientCStringFromString: (interpreterProxy stackObjectValue: 0)).
	interpreterProxy pop: 1
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 3/27/2012 21:57'!
printAllStacksFunctionAddress
	"Answer the address of the printAllStacks() function. Must be inlined in order
	to generate the expected code for sender."

	| printAllStacks |
	<inline: true>
	<returnTypeC: 'void *'>
	<var: #printAllStacks declareC: 'extern sqInt printAllStacks()'>
	^ self cCoerce: (self cCode: 'printAllStacks' inSmalltalk: [#printAllStacks]) to: 'void *'
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 1/21/2012 09:57'!
printAllStacksOnSignal: sigNum
	"Similar to #forwardSignal:toSemaphoreAt: except that the signal
	handler is set to invoke printAllStacks() on receipt of signal sigNum
	rather than signalling a Smalltalk semaphore. This is intended for
	debugging support."

	| newHandler oldHandler |
	<returnTypeC: 'void *'>
	<var: 'newHandler' type: 'void *'>
	<var: 'oldHandler' type: 'void *'>
	semaIndices = nil ifTrue: [^ nil]. "interpreter simulation"
	((semaIndices at: sigNum) > 0)
		ifTrue:
			["Handler is already set, answer an error"
			^ self sigErrorNumber].
	newHandler := self getPrintAllStacksFunction.
	oldHandler := self setSignalNumber: sigNum handler: newHandler.
	self originalSignalHandlers at: sigNum put: oldHandler.
	"Save a non-zero value in semaIndices to indicate that this signal handler is in
	effect. This permits #forwardSignal:toSemaphore: to restore the original handler."
	semaIndices at: sigNum put: -1.
	^ oldHandler
! !

!UnixOSProcessPlugin methodsFor: 'private - pthread' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/4/2006 23:20'!
resendSignal: sigNum
	"Signal sigNum has been caught by a thread other than the pthread in which
	the interpreter is executing. Rather than handling it in this thread, resend
	it to the interpreter thread."

	<var: 'sigNum' type: 'int'>
	self cCode: 'pthread_kill(vmThread, sigNum)'! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 5/9/2011 08:13'!
restoreDefaultSignalHandlers
	"Restore signal handlers to their original behaviors."

	| sigNum |
	<returnTypeC: 'void'>
	semaIndices = nil "nil if in interpreter simulation"
		ifFalse: [sigNum := 1.
			[sigNum <= self signalArraySize] whileTrue:
				[((semaIndices at: sigNum) > 0) ifTrue:
					[self setSignalNumber: sigNum handler: (self originalSignalHandlers at: sigNum).].
				sigNum := sigNum + 1]]
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 3/17/2007 17:29'!
sendSignal: sig toPid: pid
	"kill(pid, sig)"

	^ self k: pid ill: sig! !

!UnixOSProcessPlugin methodsFor: 'private - VM atexit' stamp: 'dtl 8/21/2011 22:28'!
sendSignalToPids
	"Exit function to be registered with atexit() to signal child processes on VM exit.
	Function registration is performed in initializeModuleForPlatform."

	| count pid |
	<returnTypeC: 'void'>
	<var: 'pid' type: 'pid_t'>
	count := 0.
	[count < pidCount]
		whileTrue: [pid := pidArray at: count.
			self sendSignal: sigNumToSend toPid: pid.
			count := count + 1]! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 12/22/2012 13:34'!
setSigaltstack
	"Set a new signal stack, or reuse existing signal stack if one has been previously
	allocated. Signals will be delivered on this stack to avoid consuming native stack
	space that may be required for a StackInterpreter."

	<var: 'sigstack' type: 'stack_t'> 

	self isDefined: 'SA_ONSTACK'
		inSmalltalk: []
		comment: 'true if platform supports sigaltstack'
		ifTrue: [ | sigstack SigStackSize |
			"See if there's already a sigaltstack in place"
			(self cCode: 'sigaltstack(0,&sigstack)' inSmalltalk: [0]) < 0 ifTrue:
				[self perror: 'sigaltstack'].

			self isDefined: 'SA_DISABLE'
				inSmalltalk: []
				comment: 'Use existing signal stack if available. Mac OS documents SA_DISABLE but defines SS_DISABLE'
				ifTrue: [(self cCode: 'sigstack.ss_size == 0 || (sigstack.ss_flags & SA_DISABLE)'
							inSmalltalk: [sigstack touch]) ifFalse:
						[^true]]
				ifFalse: [(self cCode: 'sigstack.ss_size == 0 || (sigstack.ss_flags & SS_DISABLE)'
							inSmalltalk: [sigstack touch]) ifFalse:
						[^true]].

			"16k words is also the amount the ITimer heartbeat routines allocate."
			SigStackSize := 1024 * (self sizeof: #'void *' asSymbol) * 16 max: #MINSIGSTKSZ asSymbol.
			nil = (self cCode: 'sigstack.ss_size = SigStackSize, sigstack.ss_sp = malloc(SigStackSize)'
					inSmalltalk: [SigStackSize touch]) ifTrue:
				[self msg: 'sigstack malloc failed'.
				 ^false].
			(self cCode: 'sigaltstack(&sigstack, 0) < 0') ifTrue:
				[self msg: 'sigaltstack install failed'.
				 self cCode: '(void)free(sigstack.ss_sp)'.
				 ^false].
			^true]
		ifFalse: ["sigaltstack is not supported on this platform. Method body is not compiled,
				and stack_t will have been redeclared as sqInt in initializeModuleForPlatform"]
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 12/22/2012 13:40'!
setSignalNumber: signalNumber handler: signalHandlerAddress
	"Set a signal handler. The C code translator will convert #sig:nal: into 'signal(parm1, parm2)'"

	| sigHandlerAction oldHandlerAction |
	<returnTypeC: 'void *'>
	<var: 'signalHandlerAddress' type: #'void *'>
	<var: 'sigHandlerAction' type: #'struct sigaction'>
	<var: 'oldHandlerAction' type: #'struct sigaction'>

	self isDefined: 'SA_ONSTACK'
		inSmalltalk: []
		comment: 'true if platform supports sigaltstack'
		ifTrue: [self cCode: 'sigHandlerAction.sa_sigaction = signalHandlerAddress'.
			self cCode: 'sigHandlerAction.sa_flags = SA_ONSTACK | SA_RESTART'.
			self cCode: 'sigemptyset(&sigHandlerAction.sa_mask)'.
			((self sig: signalNumber act: (self addressOf: sigHandlerAction) ion: (self addressOf: oldHandlerAction)) ~= 0) ifTrue:
				[self perror: 'signal'].
			^self cCode: 'oldHandlerAction.sa_sigaction'			]
		ifFalse: [^self sig: signalNumber nal: signalHandlerAddress
			"sigaction will have been defined to some random but compilable structure type
			in initializeModuleForPlatform"]
! !

!UnixOSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/3/2002 23:41'!
shutdownModule

	<export: true>
	self restoreDefaultSignalHandlers! !

!UnixOSProcessPlugin methodsFor: 'simulation support' stamp: 'eem 12/12/2012 13:05'!
sig: signum act: sigHandler ion: oldHandler
	"Simulate sigaction(3)"
	<doNotGenerate>
	^0! !

!UnixOSProcessPlugin methodsFor: 'simulation support' stamp: 'eem 6/21/2011 15:00'!
sig: signum nal: sigHandler
	"Simulate signal(3)"
	<doNotGenerate>
	^0! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/16/2012 15:36'!
sigAbrtNumber
	"Abort signal from abort(3)"

	^ self cCode: 'SIGABRT' inSmalltalk: [6]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/16/2012 15:36'!
sigAlrmNumber
	"Timer signal from alarm(2)"

	^ self cCode: 'SIGALRM' inSmalltalk: [14]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 8/3/2002 14:46'!
sigChldNumber
	"Child status has changed (POSIX)."

	^ self cCode: 'SIGCHLD' inSmalltalk: [17]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/16/2012 15:40'!
sigContNumber
	"Continue if stopped"

	^ self cCode: 'SIGCONT' inSmalltalk: [18]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl (auto pragmas dtl 2010-11-28) 2/8/2003 09:50'!
sigDefaultNumber
	"Default action for a signal"

	<returnTypeC: 'void *'>
	^ self cCode: 'SIG_DFL' inSmalltalk: [0]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl (auto pragmas dtl 2010-11-28) 2/8/2003 09:50'!
sigErrorNumber
	"Error return from signal()"

	<returnTypeC: 'void *'>
	^ self cCode: 'SIG_ERR' inSmalltalk: [-1]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/6/2002 17:47'!
sigHupNumber
	"Hangup detected on controlling terminal or death of controlling process"

	^ self cCode: 'SIGHUP' inSmalltalk: [1]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl (auto pragmas dtl 2010-11-28) 7/12/2003 11:07'!
sigIgnoreNumber
	"Ignore action for a signal"

	<returnTypeC: 'void *'>
	^ self cCode: 'SIG_IGN' inSmalltalk: [1]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 8/3/2002 14:46'!
sigIntNumber
	"Interrupt (ANSI)."

	^ self cCode: 'SIGINT' inSmalltalk: [2]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/6/2002 17:49'!
sigKillNumber
	"Kill signal"

	^ self cCode: 'SIGKILL' inSmalltalk: [9]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 8/3/2002 14:46'!
sigPipeNumber
	"Broken pipe (POSIX)."

	^ self cCode: 'SIGPIPE' inSmalltalk: [13]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/6/2002 17:48'!
sigQuitNumber
	"Quit from keyboard"

	^ self cCode: 'SIGQUIT' inSmalltalk: [3]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/16/2012 15:42'!
sigStopNumber
	"Stop process"

	^ self cCode: 'SIGSTOP' inSmalltalk: [19]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/6/2002 17:46'!
sigTermNumber
	"Termination signal. This is the default signal sent by the unix kill(1) command."

	^ self cCode: 'SIGTERM' inSmalltalk: [15]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 11/4/2005 06:39'!
sigUsr1Number
	"User defined signal number 1. This is value is platform-dependent, so the
	inSmalltalk default of 10 may be wrong on some platforms."

	^ self cCode: 'SIGUSR1' inSmalltalk: [10]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 11/4/2005 06:39'!
sigUsr2Number
	"User defined signal number 2. This is value is platform-dependent, so the
	inSmalltalk default of 12 may be wrong on some platforms."

	^ self cCode: 'SIGUSR2' inSmalltalk: [12]
! !

!UnixOSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl 1/31/2004 19:45'!
signalArraySize
	"Number of possible signals for this OS plus one. The signal handler arrays
	declared in #declareCVarsIn: are this size."

	^ self cCode: 'NSIG' inSmalltalk: [65]
! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 09:28'!
unixFileNumber: fileHandle
	"Answer the integer Unix file number corresponding to a file handle (FILE* )."

	<returnTypeC: 'int'>
	<var: 'fileHandle' type: 'FILEHANDLETYPE'>
	^ self fileno: fileHandle
! !

OSProcessPlugin subclass: #Win32OSProcessPlugin
	instanceVariableNames: 'readCharBufferArray readCharCountArray readCharStatusArray readCharArraySize readCharThreadIdArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

Win32OSProcessPlugin class
	instanceVariableNames: 'arrayValueOf'!

!Win32OSProcessPlugin commentStamp: '<historical>' prior: 0!
I provide pluggable primitive access to the Windows operating system process in which the
Squeak VM is currently running.!

Win32OSProcessPlugin class
	instanceVariableNames: 'arrayValueOf'!

!Win32OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 8/31/2008 13:36'!
declareCVarsIn: cg

	"Additional declarations are done in my subclasses. This permits a meaningful
	module identification string to be written with the name of the actual subclass
	being generated."

	cg addHeaderFile: '<sys/types.h>
/* ', self idString, ' - ', self moduleName, '.c translated from class
   ', self name, ' of OSProcessPlugin version ', self versionString, ' */'.
	cg addHeaderFile: '<windows.h>'.
	cg addHeaderFile: '<process.h>'.
	cg addHeaderFile: '"sqWin32.h"
#define FILEHANDLETYPE HANDLE	/* The type of low level stream to be used in a struct SQFile */
#define READCHARARRAYSIZE 512	/* Corresponds to expected max size of external semaphore table */
									/* Table expands dynamically, so just use a big number */'.
	cg var: 'readCharArraySize' declareC: 'int readCharArraySize = READCHARARRAYSIZE'.
	cg var: 'readCharBufferArray' declareC: 'unsigned char readCharBufferArray[READCHARARRAYSIZE]'.
	cg var: 'readCharCountArray' declareC: 'DWORD readCharCountArray[READCHARARRAYSIZE]'.
	cg var: 'readCharStatusArray' declareC: 'int readCharStatusArray[READCHARARRAYSIZE]'.
	cg var: 'readCharThreadIdArray' declareC: 'DWORD readCharThreadIdArray[READCHARARRAYSIZE]'.
! !

!Win32OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:06'!
installedModuleName
	"Answer the module name string for the plugin which is actually running in the VM"

	"Win32OSProcessPlugin installedModuleName"

	<primitive: 'primitiveModuleName' module: 'Win32OSProcessPlugin'>
	^ nil
! !

!Win32OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:24'!
installedModuleVersionString
	"Answer the version string for the plugin which is actually running in the VM"

	"Win32OSProcessPlugin installedModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'Win32OSProcessPlugin'>
	^ nil
! !

!Win32OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 10/10/2012 08:05'!
isSuitablePluginForPlatform: platformName
	"Answer true is this is an instance of the class which is responsible for
	representing the named platform. Note: for Windows, the VM reports
	platform name 'Win32', but the platforms directory name is 'win32'.
	VMMaker uses the platforms directory name, so answer true for either
	name. Newer versions of VMMaker use this mechanism"

	^ platformName asLowercase = 'win32'
! !

!Win32OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 2/21/2002 16:45'!
moduleName

	^ 'Win32OSProcessPlugin'! !

!Win32OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 8/13/2003 21:55'!
primGetPid
	"Answer the OS process ID for the OS process in which I am currently executing."

	<primitive: 'primitiveGetPid' module: 'Win32OSProcessPlugin'>
	^ nil! !

!Win32OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 8/21/2011 22:37'!
requiredMethodNames
	"return the list of method names that should be retained for export or other support reasons"

	^{ #handleAnyChildProc: .
		#readCharThread: .
		#threadMutexHandle .
		#stdinHandle .
		#stdoutHandle .
		#stderrHandle }! !

!Win32OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 7/26/2010 22:29'!
shouldBeTranslated

	^ true

! !

!Win32OSProcessPlugin methodsFor: 'private - pipes' stamp: 'dtl (auto pragmas dtl 2010-11-28) 3/11/2005 18:25'!
createPipeForReader: readerIOStreamPtr writer: writerIOStreamPtr
	"Create a pipe and populate the readerIOStream and writerIOStream variables.
	Answer true for success, else false"

	<var: 'readerIOStreamPtr' type: 'FILEHANDLETYPE *'>
	<var: 'writerIOStreamPtr' type: 'FILEHANDLETYPE *'>
	^ self cCode: 'CreatePipe(readerIOStreamPtr, writerIOStreamPtr, NULL, 0)'
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/6/2002 19:45'!
createThreadMutex
	"Answer a HANDLE for a new mutex semaphore."

	<returnTypeC: 'HANDLE'>
	^ self cCode: 'CreateMutex(NULL, false, NULL)'
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/24/2002 16:29'!
exitThreadMutexHandle
	"Answer a HANDLE for a mutex semaphore for the child handler thread."

	| h |
	<returnTypeC: 'HANDLE'>
	<var: 'h' declareC: 'static HANDLE h= 0'>
	h isNil ifTrue: [h := self createThreadMutex].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - pipes' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 20:59'!
handleAnyChildProc: args
	"Wait for a process HANDLE to exit, then signal the Semaphore at index. Answer
	the exit status of the process. This is the code to be executed in a thread to provide
	asychronous notification of an external process exit."

	| threadArgs handles count index exitStatus |
	<returnTypeC: 'DWORD WINAPI'>
	<var: 'args' type: 'void *'>
	<var: 'threadArgs' type: 'struct {int count; HANDLE *handles; int semaIndex;} *'>
	<var: 'handles' type: 'volatile LPHANDLE'>
	<var: 'count' type: 'volatile int'>
	<var: 'index' type: 'volatile int'>
	<var: 'exitStatus' type: 'DWORD'>
	threadArgs := args.
	self cCode: 'count= threadArgs->count' inSmalltalk: [0].
	self cCode: 'handles= threadArgs->handles' inSmalltalk: [0].
	self cCode: 'index= threadArgs->semaIndex' inSmalltalk: [0].

	"Release mutex, allowing additional threads to be set up"
	self releaseThreadMutex: self exitThreadMutexHandle.

	exitStatus := self cCode: 'WaitForMultipleObjects(count, handles, FALSE, INFINITE)'.
	self threadSafeSignalSemaphoreWithIndex: index.
	^ exitStatus
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:00'!
handleFrom: aHandleObject
	"Answer the HANDLE represented by aHandleObject."

	| ptr handleUnion idx |
	<returnTypeC: 'HANDLE'>
	<var: 'ptr' type: 'unsigned char *'>
	<var: 'handleUnion' type: 'union {void *address; unsigned char bytes[sizeof(void *)];}'>
	((interpreterProxy isBytes: aHandleObject) and:
			[(interpreterProxy stSizeOf: aHandleObject) = self sizeOfHandle])
		ifFalse:
			[interpreterProxy primitiveFail.
			^ nil].
	ptr := interpreterProxy arrayValueOf: aHandleObject.
	idx := 0.
	[idx < self sizeOfHandle]	whileTrue:
		[self cCode: 'handleUnion.bytes[idx] = ptr[idx]'.
		idx := idx + 1].
	^ self cCode: 'handleUnion.address' inSmalltalk: [handleUnion]
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl (auto pragmas dtl 2010-11-28) 12/26/2007 10:23'!
handleFromSQFile: anSQFileRecord

	| sqFile |
	<returnTypeC: 'HANDLE'>
	<var: 'sqFile' type: 'SQFile *'>
	((interpreterProxy isBytes: anSQFileRecord) and:
			[(interpreterProxy stSizeOf: anSQFileRecord) = self fileRecordSize])
		ifFalse:
			[interpreterProxy primitiveFail.
			^ nil].
	sqFile := self fileValueOf: anSQFileRecord.
	^ self cCode: 'sqFile->file' inSmalltalk: [0]
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:04'!
handleObjectFrom: aHandle
	"Answer an object which contains the value of a HANDLE."
 
	| size handleOop ptr handleUnion idx |
	<var: 'aHandle' type: 'HANDLE'>
	<var: 'ptr' type: 'unsigned char *'>
	<var: 'handleUnion' type: 'union {HANDLE handle; char bytes[sizeof(HANDLE)];}'>
	self cCode: 'handleUnion.handle = aHandle'.
	size := self sizeOfHandle.
	handleOop := interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: size.
	ptr := interpreterProxy arrayValueOf: handleOop.
	idx := 0.
	[idx < size] whileTrue:
		[self cCode: 'ptr[idx]= handleUnion.bytes[idx]'.
		idx := idx + 1].
	^ handleOop
! !

!Win32OSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl 3/4/2006 17:45'!
initializeModuleForPlatform
	"Platform specific initialization"! !

!Win32OSProcessPlugin methodsFor: 'primitives - console' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 07:05'!
primitiveAllocConsole
	"Allocate a console if not already allocated. Answer true on success."

	<export: true>
	self AllocConsole
		ifTrue: [interpreterProxy pop: 1 thenPush: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 1 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - nonblocking read' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:21'!
primitiveBufferValuesAt
	"For debugging only. Answer the current values of readCharBufferArray,
	readCharCountArray, and readCharStatusArray at index, an integer corresponding
	to a semaphore for one read handler thread. Answer an Array with the buffered
	character, the character count, and the status value."

	| index result byte count status |
	<export: true>
	<var: 'byte' type: 'unsigned char'>
	<var: 'count' type: 'DWORD'>
	<var: 'status' type: 'int'>
	index := interpreterProxy stackIntegerValue: 0.
	byte := readCharBufferArray at: index.
	count := readCharCountArray at: index.
	status := readCharStatusArray at: index.
	result := interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 3.
	interpreterProxy
		storePointer: 0
		ofObject: result
		withValue: (interpreterProxy positive32BitIntegerFor: byte).
	interpreterProxy
		storePointer: 1
		ofObject: result
		withValue: (interpreterProxy positive32BitIntegerFor: count).
	interpreterProxy
		storePointer: 2
		ofObject: result
		withValue: (interpreterProxy positive32BitIntegerFor: status).
	interpreterProxy pop: 2 thenPush: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:21'!
primitiveCanAccessChildProcess
	"Answer true if the OS process represented by a HANDLE can be accessed by this OS process."

	| handle exitStatus |
	<export: true>
	<var: 'handle' type: 'HANDLE'>
	<var: 'exitStatus' type: 'DWORD'>
	(handle := self handleFrom: (interpreterProxy stackObjectValue: 0)) isNil ifTrue: [^ nil].
	(((self cCode: 'GetExitCodeProcess(handle, &exitStatus)' inSmalltalk: [0]) ~= 0)
		and: [self cCode: 'exitStatus == STILL_ACTIVE' inSmalltalk: [false]])
			ifTrue: [interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
			ifFalse: [interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]! !

!Win32OSProcessPlugin methodsFor: 'primitives - handles' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:23'!
primitiveCloseHandle
	"Close the specified handle, which may refer to a process, a thread, or
	some other Win32 object."

	| handleOop handle result |
	<export: true>
	<var: 'handle' type: 'HANDLE'>
	<var: 'result' type: 'BOOL'>
	handleOop := interpreterProxy stackObjectValue: 0.
	(handle := self handleFrom: handleOop) isNil ifTrue: [^ nil].
	result := self CloseHandle: handle.
	interpreterProxy pop: 2.
	(result ~= 0)
		ifTrue: [interpreterProxy push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy push: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:06'!
primitiveCommand
	"Create a process to run commandString. Answer an Array formed from the
	PROCESS_INFORMATION structure for the new process. The caller is expected
	to close the process and thread handles when they are no longer needed.
	The result array contains hProcess, hThread, dwProcessId, dwThreadId.

	This primitive has been replaced by #primitiveCommandWithInputOutputError,
	and will be removed in future versions of OSProcess."

	| cStringPtr commandString startUp procInfo processInformation |
	<export: true>
	<var: 'cStringPtr' type: 'LPTSTR'>
	<var: 'startUp' type: 'STARTUPINFO'>
	<var: 'procInfo' type: 'PROCESS_INFORMATION'>

	commandString := interpreterProxy stackObjectValue: 0.
	cStringPtr := self cStringFromString: commandString.

	self cCode: 'GetStartupInfo (&startUp)'.
	self cCode: 'startUp.hStdInput = GetStdHandle(STD_INPUT_HANDLE)'.
	self cCode: 'startUp.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE)'.
	self cCode: 'startUp.hStdError = GetStdHandle(STD_ERROR_HANDLE)'.
	(self cCode:  'CreateProcess(NULL, cStringPtr, NULL, NULL, TRUE, NORMAL_PRIORITY_CLASS, NULL, NULL, &startUp, &procInfo)')
		ifFalse: [^ interpreterProxy primitiveFail].

	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: (self cCode: 'procInfo.dwThreadId')).
	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: (self cCode: 'procInfo.dwProcessId')).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: (self cCode: 'procInfo.hThread')).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: (self cCode: 'procInfo.hProcess')).
	processInformation := interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 4.
	interpreterProxy
		storePointer: 0
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 1
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 2
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 3
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy pop: 2; push: processInformation
! !

!Win32OSProcessPlugin methodsFor: 'primitives - process creation' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:24'!
primitiveCommandWithInputOutputError
	"Create a process to run commandString, using the specified input, output,
	and error streams. If any of the input, output, or error streams is nil, then
	the current values of these handles in this OS process will be passed to the
	new process. Answer an Array formed from the PROCESS_INFORMATION
	structure for the new process. The caller is expected to close the process and
	thread handles when they are no longer needed. The result array contains
	hProcess, hThread, dwProcessId, dwThreadId."

	| cStringPtr commandString startUp procInfo processInformation stdErr stdOut stdIn
	inputHandle outputHandle errorHandle |
	<export: true>
	<var: 'cStringPtr' type: 'LPTSTR'>
	<var: 'startUp' type: 'STARTUPINFO'>
	<var: 'procInfo' type: 'PROCESS_INFORMATION'>
	<var: 'inputHandle' type: 'HANDLE'>
	<var: 'outputHandle' type: 'HANDLE'>
	<var: 'errorHandle' type: 'HANDLE'>
	stdErr := interpreterProxy stackObjectValue: 0.
	stdOut := interpreterProxy stackObjectValue: 1.
	stdIn := interpreterProxy stackObjectValue: 2.
	commandString := interpreterProxy stackObjectValue: 3.
	cStringPtr := self cStringFromString: commandString.
	(stdIn = interpreterProxy nilObject)
		ifTrue: [inputHandle := self stdInHandle]
		ifFalse: [inputHandle := self handleFromSQFile: stdIn].
	(stdOut = interpreterProxy nilObject)
		ifTrue: [outputHandle := self stdOutHandle]
		ifFalse: [outputHandle := self handleFromSQFile: stdOut].
	(stdErr = interpreterProxy nilObject)
		ifTrue: [errorHandle := self stdErrHandle]
		ifFalse: [errorHandle := self handleFromSQFile: stdErr].
	self cCode: 'GetStartupInfo (&startUp)'.
	self cCode: 'startUp.hStdInput = inputHandle'.
	self cCode: 'startUp.hStdOutput = outputHandle'.
	self cCode: 'startUp.hStdError = errorHandle'.

	(self cCode:  'CreateProcess(NULL, cStringPtr, NULL, NULL, TRUE, NORMAL_PRIORITY_CLASS, NULL, NULL, &startUp, &procInfo)')
		ifFalse: [^ interpreterProxy primitiveFail].

	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: (self cCode: 'procInfo.dwThreadId')).
	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: (self cCode: 'procInfo.dwProcessId')).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: (self cCode: 'procInfo.hThread')).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: (self cCode: 'procInfo.hProcess')).
	processInformation := interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 4.
	interpreterProxy
		storePointer: 0
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 1
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 2
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 3
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy pop: 5; push: processInformation
! !

!Win32OSProcessPlugin methodsFor: 'primitives - console' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 07:04'!
primitiveFreeConsole
	"Deallocate the console if allocated. Answer true on success."

	<export: true>
	self FreeConsole
		ifTrue: [interpreterProxy pop: 1 thenPush: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 1 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 9/10/2011 21:02'!
primitiveGetCurrentWorkingDirectory
	"Answer a string containing the current working directory."

	| buffer requiredLen cwdLen cwdString |
	<export: true>
	<var: 'buffer' type: 'char *'>
	<var: 'requiredLen' type: 'DWORD'>
	<var: 'cwdLen' type: 'DWORD'>

	"Allocate a result buffer"
	requiredLen := self cCode: 'GetCurrentDirectory(0, NULL)' inSmalltalk: [0].
	(requiredLen = 0) ifTrue: [^ interpreterProxy primitiveFail].
	buffer := self cCode: 'alloca(requiredLen * sizeof(TCHAR))'.

	"Call GetCurrentDirectory() to obtain null-terminated result string"
	cwdLen := self cCode: 'GetCurrentDirectory(requiredLen, buffer)' inSmalltalk: [0].
	(cwdLen = 0) ifTrue: [^ interpreterProxy primitiveFail].

	"Answer a String with null terminator removed"
	cwdString := self stringFromCString: buffer.
	interpreterProxy pop: 1thenPush: cwdString
 ! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:25'!
primitiveGetEnvironmentStrings
	"Answer the environment block in the form of an Array of Strings. The
	caller is expected to parse the strings into a dictionary of keys and values."

	| pVarBlock p envArray count string |
	<export: true>
	<var: 'pVarBlock' type: 'TCHAR *'>
	<var: 'p' type: 'TCHAR *'>
	pVarBlock := self GetEnvironmentStrings.

	"Count the environment strings so the result array can be pre-allocated.
	This minimizes the number of objects which must be pushed onto the
	stack of remappable objects."
	p := pVarBlock.
	count := 0.
	[self cCode: '*p !!= 0' inSmalltalk: [pVarBlock ~= nil]] whileTrue:
		[count := count + 1.
		[self cCode: '*p !!= 0' inSmalltalk: [p ~= nil]] whileTrue: [p := p + 1].
		p := p + 1].
	"Allocate an array for the strings."
	envArray := interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: count.
	"Populate the array."
	p := pVarBlock.
	count := 0.
	[self cCode: '*p !!= 0' inSmalltalk: [pVarBlock ~= nil]] whileTrue:
		[count := count + 1.
		interpreterProxy pushRemappableOop: envArray.
		string := self stringFromCString: p.	"May trigger garbage collection"
		envArray := interpreterProxy popRemappableOop.
		interpreterProxy stObject: envArray at: count put: string.
		[self cCode: '*p !!= 0' inSmalltalk: [p ~= nil]] whileTrue: [p := p + 1].
		p := p + 1].
	self FreeEnvironmentStrings: pVarBlock.
	interpreterProxy pop: 1 thenPush: envArray! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:22'!
primitiveGetExitStatusForHandle
	"Answer the exit status for the process represented by a HANDLE. Fail if
	the process is still active, or if the GetExitCodeProcess call fails."

	| handle exitStatus |
	<export: true>
	<var: 'handle' type: 'HANDLE'>
	<var: 'exitStatus' type: 'DWORD'>
	(handle := self handleFrom: (interpreterProxy stackObjectValue: 0)) isNil ifTrue: [^ nil].
	((self cCode: 'GetExitCodeProcess(handle, &exitStatus)' inSmalltalk: [0]) ~= 0)
		ifTrue:
			[(self cCode: 'exitStatus == STILL_ACTIVE' inSmalltalk: [false]) ifTrue:
				["Child process is still active, fail the primitive"
				^ interpreterProxy primitiveFail].
			interpreterProxy
				pop: 2
				thenPush: (interpreterProxy positive32BitIntegerFor: exitStatus)]
		ifFalse:
			[^ interpreterProxy primitiveFail]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 2/27/2002 12:40'!
primitiveGetMainThreadHandle
	"Answer the ID of my main thread."

	| handle handleOop |
	<export: true>
	<var: 'handle' type: 'HANDLE'>
	handle := self GetCurrentThread.
	handleOop := self handleObjectFrom: handle.
	interpreterProxy pop: 1 thenPush: handleOop
! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:25'!
primitiveGetMainThreadID
	"Answer the ID of my main thread."

	| tid tidOop |
	<export: true>
	<var: 'tid' type: 'DWORD'>
	tid := self GetCurrentThreadId.
	tidOop := interpreterProxy positive32BitIntegerFor: tid.
	interpreterProxy pop: 1 thenPush: tidOop
! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:26'!
primitiveGetPid
	"Answer the process ID of my OS process"

	| pid pidOop |
	<export: true>
	<var: 'pid' type: 'DWORD'>
	pid := self GetCurrentProcessId.
	pidOop := interpreterProxy positive32BitIntegerFor: pid.
	interpreterProxy pop: 1 thenPush: pidOop
! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 2/27/2002 12:40'!
primitiveGetPidHandle
	"Answer the handle for my OS process"

	| handle handleOop |
	<export: true>
	<var: 'handle' type: 'HANDLE'>
	handle := self GetCurrentProcess.
	handleOop := self handleObjectFrom: handle.
	interpreterProxy pop: 1 thenPush: handleOop
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:26'!
primitiveGetStdError
	"Answer a two element array containing the sqFile data structure representing
	standard error stream for my OS process, and a flag (true or false) to indicate
	whether the sqFile data structure contains a valid HANDLE. If no standard error
	stream is available for this OS process, the sqFile data structure will contain an
	invalid HANDLE value, which will result in failures on subsequent accesses."

	| file handle result invalid thisSession fileOop |
	<export: true>
	<var: 'file' type: 'SQFile *'>
	<var: 'handle' type: 'HANDLE'>
	<var: 'invalid' type: 'HANDLE'>
	thisSession := self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	invalid := self cCode: 'INVALID_HANDLE_VALUE' inSmalltalk: [-1].
	result := interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 2.
	fileOop := self newSQFileByteArray.
	file := self fileValueOf: fileOop.
	handle := self cCode: 'GetStdHandle(STD_ERROR_HANDLE)' inSmalltalk: [-1].
	(handle ~= invalid)
		ifTrue: [interpreterProxy stObject: result at: 2 put: interpreterProxy trueObject]
		ifFalse: [interpreterProxy stObject: result at: 2 put: interpreterProxy falseObject].
	self cCode: 'file->file = handle'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy stObject: result at: 1 put: fileOop.
	interpreterProxy pop: 2; push: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:27'!
primitiveGetStdInput
	"Answer a two element array containing the sqFile data structure representing
	standard input stream for my OS process, and a flag (true or false) to indicate
	whether the sqFile data structure contains a valid HANDLE. If no standard input
	stream is available for this OS process, the sqFile data structure will contain an
	invalid HANDLE value, which will result in failures on subsequent accesses."

	| file handle result invalid thisSession fileOop |
	<export: true>
	<var: 'file' type: 'SQFile *'>
	<var: 'handle' type: 'HANDLE'>
	<var: 'invalid' type: 'HANDLE'>
	thisSession := self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	invalid := self cCode: 'INVALID_HANDLE_VALUE' inSmalltalk: [-1].
	result := interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 2.
	fileOop := self newSQFileByteArray.
	file := self fileValueOf: fileOop.
	handle := self cCode: 'GetStdHandle(STD_INPUT_HANDLE)' inSmalltalk: [-1].
	(handle ~= invalid)
		ifTrue: [interpreterProxy stObject: result at: 2 put: interpreterProxy trueObject]
		ifFalse: [interpreterProxy stObject: result at: 2 put: interpreterProxy falseObject].
	self cCode: 'file->file = handle'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 0'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy stObject: result at: 1 put: fileOop.
	interpreterProxy pop: 2; push: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:27'!
primitiveGetStdOutput
	"Answer a two element array containing the sqFile data structure representing
	standard output stream for my OS process, and a flag (true or false) to indicate
	whether the sqFile data structure contains a valid HANDLE. If no standard output
	stream is available for this OS process, the sqFile data structure will contain an
	invalid HANDLE value, which will result in failures on subsequent accesses."

	| file handle result invalid thisSession fileOop |
	<export: true>
	<var: 'file' type: 'SQFile *'>
	<var: 'handle' type: 'HANDLE'>
	<var: 'invalid' type: 'HANDLE'>
	thisSession := self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	invalid := self cCode: 'INVALID_HANDLE_VALUE' inSmalltalk: [-1].
	result := interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 2.
	fileOop := self newSQFileByteArray.
	file := self fileValueOf: fileOop.
	handle := self cCode: 'GetStdHandle(STD_OUTPUT_HANDLE)' inSmalltalk: [-1].
	(handle ~= invalid)
		ifTrue: [interpreterProxy stObject: result at: 2 put: interpreterProxy trueObject]
		ifFalse: [interpreterProxy stObject: result at: 2 put: interpreterProxy falseObject].
	self cCode: 'file->file = handle'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy stObject: result at: 1 put: fileOop.
	interpreterProxy pop: 2; push: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - nonblocking read' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/1/2002 19:39'!
primitiveLastReadFor
	"A character has been read into an external buffer corresponding to aSemaphoreIndex,
	and is now available. Answer the integer value of the character, or nil if no character
	was read, or -1 if an error occurred on the read."

	| index result |
	<export: true>
	index := interpreterProxy stackIntegerValue: 0.
	(readCharStatusArray at: index)
		ifFalse:
			[result := interpreterProxy integerObjectOf: -1]
		ifTrue:
			[((readCharCountArray at: index) < 1)
				ifTrue:
					[result := interpreterProxy nilObject]
				ifFalse:
					[result := interpreterProxy integerObjectOf: (readCharBufferArray at: index)]].
	interpreterProxy pop: 2 thenPush: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - nonblocking read' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/28/2002 07:42'!
primitiveLastReadForStoreIn
	"A character has been read into an external buffer corresponding to aSemaphoreIndex,
	and is now available. Answer the integer value of the character, or nil if no character
	was read, or -1 if an error occurred on the read. Store the full results of the most
	recent read into a three element array provided by the sender. Contents of the array
	will be status of the read call, character read, and character count (which should
	always be 1)."

	| index resultArray result |
	<export: true>
	index := interpreterProxy stackIntegerValue: 1.
	resultArray := interpreterProxy stackValue: 0.
	interpreterProxy stObject: resultArray at: 1 put:
		(interpreterProxy positive32BitIntegerFor: (readCharStatusArray at: index)).
	interpreterProxy stObject: resultArray at: 2 put:
		(interpreterProxy integerObjectOf: (readCharBufferArray at: index)).
	interpreterProxy stObject: resultArray at: 3 put:
		(interpreterProxy integerObjectOf: (readCharCountArray at: index)).
	(readCharStatusArray at: index)
		ifFalse:
			[result := interpreterProxy integerObjectOf: -1]
		ifTrue:
			[((readCharCountArray at: index) < 1)
				ifTrue:
					[result := interpreterProxy nilObject]
				ifFalse:
					[result := interpreterProxy integerObjectOf: (readCharBufferArray at: index)]].
	interpreterProxy pop: 3 thenPush: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - nonblocking read' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/18/2002 09:39'!
primitiveNextFromSignaling
	"Read the next character from anIOHandle (a SQFile struct) into a buffer in the VM. When
	the read completes, signal the specified Semaphore to notify that the character is available.
	Answer the handle of the new thread, or nil on error."

	| index handle hThread threadArgs threadID |
	<export: true>
	<var: 'index' type: 'static int'>
	<var: 'handle' type: 'HANDLE'>
	<var: 'hThread' type: 'HANDLE'>
	<var: 'threadID' declareC: 'DWORD threadID'>
	<var: 'threadArgs' declareC: 'static struct {int index; HANDLE handle;} threadArgs'>

	index := interpreterProxy stackIntegerValue: 0.
	(index >= readCharArraySize) ifTrue: [^ interpreterProxy primitiveFail].

	"Complete setup of one handler before doing the next. Mutex will be
	released in the newly started thread."
	self waitForThreadMutex: self readThreadMutexHandle.

	handle := self handleFromSQFile: (interpreterProxy stackObjectValue: 1).
	handle isNil ifTrue: [^ nil]. "Fail if not a handle, probably due to file having been closed"
	readCharBufferArray at: index put: 0.
	readCharCountArray at: index put: 0.
	readCharStatusArray at: index put: 0.
	self cCode: 'threadArgs.index = index'.
	self cCode: 'threadArgs.handle = handle'.
	hThread := self cCode: 'CreateThread(NULL, 1024, readCharThread, &threadArgs, 0, &threadID)'.
	"Safety net: Remember the identity of the thread which should be active for this
	semaphore index."
	readCharThreadIdArray at: index put: threadID.
	hThread isNil ifTrue: [^ interpreterProxy nilObject].
	interpreterProxy pop: 3 thenPush: (self handleObjectFrom: hThread)
! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'dtl 10/10/2012 08:05'!
primitiveOneShot
	"Answer true the first time this primitive is called in a Squeak session, and false thereafter."

	| thisPrimHasBeenCalled |
	<export: true>
	<var: 'thisPrimHasBeenCalled' declareC: 'static int thisPrimHasBeenCalled= 0'>
	(thisPrimHasBeenCalled = 0)
		ifTrue:
			[thisPrimHasBeenCalled := 1.
			interpreterProxy pop: 1 thenPush: interpreterProxy trueObject]
		ifFalse:
			[interpreterProxy pop: 1 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:27'!
primitiveSetStdErr
	"Set the file handle for standard error of my OS process to the handle passed in a SQFile struct."

	| sqFileOop handle |
	<export: true>
	<var: 'sqFile' type: 'SQFile *'>
	<var: 'handle' type: 'HANDLE'>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	handle := self fileHandleFrom: sqFileOop.
	(self cCode: 'SetStdHandle(STD_ERROR_HANDLE, handle)' inSmalltalk: [false])
		ifTrue: [interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:27'!
primitiveSetStdIn
	"Set the file handle for standard input of my OS process to the handle passed in a SQFile struct."

	| sqFileOop handle |
	<export: true>
	<var: 'sqFile' type: 'SQFile *'>
	<var: 'handle' type: 'HANDLE'>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	handle := self fileHandleFrom: sqFileOop.
	(self cCode: 'SetStdHandle(STD_INPUT_HANDLE, handle)' inSmalltalk: [false])
		ifTrue: [interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:28'!
primitiveSetStdOut
	"Set the file handle for standard output of my OS process to the handle passed in a SQFile struct."

	| sqFileOop handle |
	<export: true>
	<var: 'sqFile' type: 'SQFile *'>
	<var: 'handle' type: 'HANDLE'>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	handle := self fileHandleFrom: sqFileOop.
	(self cCode: 'SetStdHandle(STD_OUTPUT_HANDLE, handle)' inSmalltalk: [false])
		ifTrue: [interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'nice 11/13/2016 19:28'!
primitiveSetWaitForAnyProcessExitThenSignalSemaphoreWithIndex
	"Set up a thread to wait for a process HANDLE to exit, then signal the Semaphore
	at index. This provides asychronous notification of an external process exit. Answer
	the handle of the handler thread."

	| handleArrayOop handleArrayOopPtr handleArray handleArrayPtr handleCount arraySize idx semaIndex threadInformation |
	<export: true>
	<var: 'handleArrayOopPtr' type: 'sqInt *'>
	<var: 'handleArray' declareC: 'static HANDLE handleArray[MAXIMUM_WAIT_OBJECTS]'>
	<var: 'handleArrayPtr' type: 'static HANDLE *'>
	semaIndex := interpreterProxy stackIntegerValue: 0.
	handleArrayOop := interpreterProxy stackObjectValue: 1.
	handleArrayOopPtr := interpreterProxy firstIndexableField: handleArrayOop.
	handleCount := interpreterProxy stSizeOf: handleArrayOop.
	arraySize := self cCode: 'MAXIMUM_WAIT_OBJECTS' inSmalltalk: [64].
	(handleCount > arraySize) ifTrue: [^ interpreterProxy primitiveFail].
	idx := 0.
	[idx < handleCount] whileTrue:
		[handleArray at: idx put: (self handleFrom: (handleArrayOopPtr at: idx)).
		idx := idx + 1].
	handleArrayPtr := handleArray.
	threadInformation := self startHandlerThreadForAny: handleArrayPtr
							count: handleCount
							signalingSemaphoreAt: semaIndex.
	interpreterProxy pop: 3 thenPush: threadInformation
! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:23'!
primitiveTerminateThread
	"Kill the thread. No cleanup is performed, so use with caution for a thread which
	is (for example) manipulating a mutex. Answer true for success, else false."

	| handle |
	<export: true>
	<var: 'handle' type: 'HANDLE'>
	<var: 'exitStatus' type: 'DWORD'>
	(handle := self handleFrom: (interpreterProxy stackObjectValue: 0)) isNil ifTrue: [^ nil].
	(self cCode: 'TerminateThread(handle, 0)' inSmalltalk: [false])
		ifTrue: [^ interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
		ifFalse: [^ interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'private - nonblocking read thread' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/24/2002 10:09'!
readCharThread: args
	"This is the entry point for the thread procedure for reading one character from
	a file or pipe handle. There may be many instances of this thread procedure, in
	which case each thread may be distinguished by the index value with which it is
	called. Remember the index value by saving it in a variable on the thread stack.
	Read one character, and signal a Smalltalk Semaphore to indicate completion."

	| index handle threadArgs |
	<returnTypeC: 'DWORD WINAPI'>
	<var: 'handle' type: 'HANDLE'>
	<var: 'args' type: 'void *'>
	<var: 'threadArgs' type: 'struct {int index; HANDLE handle;} *'>
	threadArgs := args.
	self cCode: 'index = threadArgs->index'.
	self cCode: 'handle = threadArgs->handle'.
	"Save arguments in stack local variables. The mutex semaphore will be
	released as soon as these variables are safely on a local thread stack."
	^ self readCharThreadIndex: index handle: handle
! !

!Win32OSProcessPlugin methodsFor: 'private - nonblocking read thread' stamp: 'eem 9/26/2009 11:48'!
readCharThreadIndex: index handle: handle
	"This is the thread procedure for reading one character from a file or pipe
	handle. The values of index and handle have been passed on the local
	thread stack, so they are now local to the thread which is executing this
	method. Once the values are safely on a local thread stack, the mutex
	semaphore is released, allowing other threads to be created."

	| status |
	<returnTypeC: 'DWORD WINAPI'>
	<var: 'handle' type: 'HANDLE'>
	"Release mutex, allowing additional threads to be set up"
	self releaseThreadMutex: self readThreadMutexHandle.
	"Read a character, possibly waiting indefinitely"
	status := self cCode: 'ReadFile(handle, &readCharBufferArray[index], 1, &readCharCountArray[index], NULL)'.
	readCharStatusArray at: index put: status.
	"Ensure that this thread is still valid for this semaphore index."
	self GetCurrentThreadId = (readCharThreadIdArray at: index)
		ifTrue:
			["Signal Squeak that the character is now available"
			interpreterProxy signalSemaphoreWithIndex: index]
		ifFalse:
			["Invalidate any data we just placed in the buffer"
			readCharBufferArray at: index put: 0.
			readCharCountArray at: index put: 0.
			readCharStatusArray at: index put: 0].
	^ status
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/25/2002 11:07'!
readThreadMutexHandle
	"Answer a HANDLE for a mutex semaphore for sychronizing setup of read threads."

	| h |
	<returnTypeC: 'HANDLE'>
	<var: 'h' declareC: 'static HANDLE h= 0'>
	h isNil ifTrue: [h := self createThreadMutex].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/24/2002 10:56'!
releaseThreadMutex: aHandle
	"To be called at the end of a critical section."

	<returnTypeC: 'BOOL'>
	<var: 'aHandle' type: 'HANDLE'>
	^ self ReleaseMutex: aHandle
! !

!Win32OSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/6/2003 05:16'!
shutdownModule

	<export: true>
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/25/2002 13:22'!
signalSemaphoreMutexHandle
	"Answer a HANDLE for a mutex semaphore for event notification."

	| h |
	<returnTypeC: 'HANDLE'>
	<var: 'h' declareC: 'static HANDLE h= 0'>
	h isNil ifTrue: [h := self createThreadMutex].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl 2/23/2002 13:35'!
sizeOfHandle

	^ self cCode: 'sizeof(HANDLE)' inSmalltalk: [4]
! !

!Win32OSProcessPlugin methodsFor: 'private - exit handler thread' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/2/2008 21:03'!
startHandlerThreadForAny: handleArrayPointer count: arraySize signalingSemaphoreAt: semaphoreIndex
	"Wait for a process HANDLE to exit, then signal the Semaphore at index. Answer
	the thread ID of the handler. This is the code to be executed in a thread to provide
	asychronous notification of an external process exit."

	| threadID hThread threadInformation threadArgs |
	<var: 'handleArrayPointer' type: 'HANDLE *'>
	<var: 'threadID' type: 'DWORD'>
	<var: 'hThread' type: 'HANDLE'>
	<var: 'threadArgs' type: 'static struct {int count; HANDLE *handles; int semaIndex;}'>

	"Complete setup of one handler before doing the next. Mutex will be
	released in the newly started thread."
	self waitForThreadMutex: self exitThreadMutexHandle.

	"Set up the thread argument structure."
	self cCode: 'threadArgs.count= arraySize'.
	self cCode: 'threadArgs.handles= handleArrayPointer'.
	self cCode: 'threadArgs.semaIndex= semaphoreIndex'.

	hThread := self cCode: 'CreateThread(NULL, 1024, handleAnyChildProc, &threadArgs, 0, &threadID)'.
	hThread isNil ifTrue: [^ interpreterProxy nilObject].
	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: threadID).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: hThread).
	threadInformation := interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 2.
	interpreterProxy
		storePointer: 0
		ofObject: threadInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 1
		ofObject: threadInformation
		withValue: interpreterProxy popRemappableOop.
	^ threadInformation
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/8/2002 22:23'!
stdErrHandle
	"Answer the pseudo HANDLE for standard error."

	<returnTypeC: 'HANDLE'>
	^ self cCode: '(HANDLE) STD_ERROR_HANDLE' inSmalltalk: ['16rFFFFFFF4' asNumber]
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/8/2002 22:23'!
stdInHandle
	"Answer the pseudo HANDLE for standard input."

	<returnTypeC: 'HANDLE'>
	^ self cCode: '(HANDLE) STD_INPUT_HANDLE' inSmalltalk: ['16rFFFFFFF6' asNumber]
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/8/2002 22:23'!
stdOutHandle
	"Answer the pseudo HANDLE for standard output."

	<returnTypeC: 'HANDLE'>
	^ self cCode: '(HANDLE) STD_OUTPUT_HANDLE' inSmalltalk: ['16rFFFFFFF5' asNumber]
! !

!Win32OSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:41'!
stderrHandle

	| h |
	<returnTypeC: 'HANDLE'>
	<var: 'h' type: 'HANDLE'>
	h := self cCode: 'GetStdHandle(STD_ERROR_HANDLE)' inSmalltalk: [0].
	(self cCode: 'h == INVALID_HANDLE_VALUE' inSmalltalk: [-1])
		ifTrue: [interpreterProxy primitiveFail].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:38'!
stdinHandle

	| h |
	<returnTypeC: 'HANDLE'>
	<var: 'h' type: 'HANDLE'>
	h := self cCode: 'GetStdHandle(STD_INPUT_HANDLE)' inSmalltalk: [0].
	(self cCode: 'h == INVALID_HANDLE_VALUE' inSmalltalk: [-1])
		ifTrue: [interpreterProxy primitiveFail].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl (auto pragmas dtl 2010-11-28) 11/19/2006 16:42'!
stdoutHandle

	| h |
	<returnTypeC: 'HANDLE'>
	<var: 'h' type: 'HANDLE'>
	h := self cCode: 'GetStdHandle(STD_OUTPUT_HANDLE)' inSmalltalk: [0].
	(self cCode: 'h == INVALID_HANDLE_VALUE' inSmalltalk: [-1])
		ifTrue: [interpreterProxy primitiveFail].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - exit handler thread' stamp: 'dtl (auto pragmas dtl 2010-11-28) 2/26/2002 14:17'!
threadMutexHandle
	"Answer a HANDLE for a mutex semaphore."

	| h |
	<returnTypeC: 'HANDLE'>
	<var: 'h' declareC: 'static HANDLE h= 0'>
	(h = 0) ifTrue: [h := self cCode: 'CreateMutex(NULL, false, NULL)'].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl 8/25/2002 13:25'!
threadSafeSignalSemaphoreWithIndex: index

	"This method should call synchronizedSignalSemaphoreWithIndex() in
	sqWin32Window.c in order to ensure that multiple threads do not execute
	vm code concurrently. Unfortunately (FIXME) I do not know how to call
	synchronizedSignalSemaphoreWithIndex() from a DLL, so this will have to
	do for now."

	self waitForThreadMutex: self signalSemaphoreMutexHandle.
	interpreterProxy signalSemaphoreWithIndex: index.
	self releaseThreadMutex: self signalSemaphoreMutexHandle
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl (auto pragmas dtl 2010-11-28) 8/24/2002 10:56'!
waitForThreadMutex: aHandle
	"To be called before entering a critical section."

	<returnTypeC: 'DWORD'>
	<var: 'aHandle' type: 'HANDLE'>
	^ self cCode: 'WaitForSingleObject(aHandle, INFINITE)'
! !
